"""Temperature profile loaded from file."""
import typing as t

import numpy as np

from taurex.types import PathLike
from taurex.util import conversion_factor

from .temparray import TemperatureArray


class TemperatureFile(TemperatureArray):
    """A temperature profile read from file.

    If pressure is included in file then the temperature profile
    will be interpolated to the pressure points of the atmosphere.

    """

    def __init__(
        self,
        filename: t.Optional[PathLike] = None,
        skiprows: t.Optional[int] = 0,
        temp_col: t.Optional[int] = 0,
        press_col: t.Optional[int] = None,
        temp_units: t.Optional[str] = "K",
        press_units: t.Optional[str] = "Pa",
        delimiter: t.Optional[str] = None,
        reverse: t.Optional[bool] = False,
    ):
        """Initialize temperature profile from file.

        Parameters
        ----------
        filename : str
            File name for temperature profile
        skiprows : int, optional
            Number of rows to skip
        temp_col : int, optional
            Column number for temperature
        press_col : int, optional
            Column number for pressure
        temp_units : str, optional
            Temperature units
        press_units : str, optional
            Pressure units
        delimiter : str, optional
            Delimiter
        reverse : bool, optional
            Reverse the order of the array

        """
        pressure_arr = None
        temperature_arr = None

        convert_t = conversion_factor(temp_units, "K")
        convert_p = conversion_factor(press_units, "Pa")

        if press_col is not None:
            arr = np.loadtxt(
                filename,
                skiprows=skiprows,
                usecols=(int(press_col), int(temp_col)),
                delimiter=delimiter,
            )
            temperature_arr = arr[:, 1] * convert_t
            pressure_arr = arr[:, 0] * convert_p
        else:
            arr = np.loadtxt(
                filename,
                skiprows=skiprows,
                usecols=int(temp_col),
            )
            temperature_arr = arr[:] * convert_t

        super().__init__(tp_array=temperature_arr, p_points=pressure_arr)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        return (
            "file",
            "fromfile",
        )



"""Modules dealing with reading data from observations."""

from .array import ArraySpectrum
from .observed import ObservedSpectrum
from .spectrum import BaseSpectrum
from .taurex import TaurexSpectrum

__all__ = ["BaseSpectrum", "ObservedSpectrum", "ArraySpectrum", "TaurexSpectrum"]



"""Base contribution classes and functions for computing optical depth."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.citation import Citable
from taurex.data.fittable import Fittable
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable

if t.TYPE_CHECKING:
    from taurex.model.model import ForwardModel
else:
    ForwardModel = object


contribute_tau: t.Callable[
    [
        int,
        int,
        int,
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        int,
        int,
        int,
        npt.NDArray[np.float64],
    ],
    npt.NDArray[np.float64],
] = None


def contribute_tau_numpy(
    startk: int,
    endk: int,
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    nlayers: int,
    ngrid: int,
    layer: int,
    tau: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Generic cross-section integration function for tau

    This has the form:

    .. math::

        \\tau_{\\lambda}(z) = \\int_{z_{0}}^{z_{1}} \\sigma(z') \\rho(z') dz',

    where :math:`z` is the layer, :math:`z_0` and :math:`z_1` are ``startK``
    and ``endK`` respectively. :math:`\\sigma` is the weighted
    cross-section ``sigma``. :math:`rho` is the ``density`` and
    :math:`dz'` is the integration path length ``path``


    Parameters
    ----------
    startK: int
        starting layer in integration

    endK: int
        last layer in integration

    density_offset: int
        Which part of the density profile to start from

    sigma: :obj:`array`
        cross-section

    density: array_like
        density profile of atmosphere

    path: array_like
        path-length or altitude gradient

    nlayers: int
        Total number of layers (unused)

    ngrid: int
        total number of grid points

    layer: int
        Which layer we currently on

    Returns
    -------
    tau : array_like
        optical depth (well almost you still need to do ``exp(-tau)`` yourself)

    """
    _path = path[startk:endk, None]
    _density = density[startk + density_offset : endk + density_offset, None]
    _sigma = sigma[startk + layer : endk + layer, :]

    tau[layer, :] += np.sum(_sigma * _path * _density, axis=0)

    return tau


def contribute_tau_numba(
    startK: int,  # noqa: N803
    endK: int,  # noqa: N803
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    nlayers: int,
    ngrid: int,
    layer: int,
    tau: npt.NDArray[np.float64],
):
    """Generic cross-section integration function for tau

    numba-fied for performance.

    This has the form:

    .. math::

        \\tau_{\\lambda}(z) = \\int_{z_{0}}^{z_{1}} \\sigma(z') \\rho(z') dz',

    where :math:`z` is the layer, :math:`z_0` and :math:`z_1` are ``startK``
    and ``endK`` respectively. :math:`\\sigma` is the weighted
    cross-section ``sigma``. :math:`rho` is the ``density`` and
    :math:`dz'` is the integration path length ``path``


    Parameters
    ----------
    startK: int
        starting layer in integration

    endK: int
        last layer in integration

    density_offset: int
        Which part of the density profile to start from

    sigma: :obj:`array`
        cross-section

    density: array_like
        density profile of atmosphere

    path: array_like
        path-length or altitude gradient

    nlayers: int
        Total number of layers (unused)

    ngrid: int
        total number of grid points

    layer: int
        Which layer we currently on

    Returns
    -------
    tau : array_like
        optical depth (well almost you still need to do ``exp(-tau)`` yourself)

    """
    for k in range(startK, endK):
        _path = path[k]
        _density = density[k + density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            tau[layer, wn] += sigma[k + layer, wn] * _path * _density


try:
    import numba

    contribute_tau = numba.jit(contribute_tau_numba, nopython=True, nogil=True)

except ImportError:
    contribute_tau = contribute_tau_numpy


class Contribution(Fittable, Logger, Writeable, Citable):
    """The base class for modelling contributions to the optical depth.

    *Abstract class*

    By default this handles contributions from cross-sections.
    If the type of contribution being implemented is a `sigma`-type
    like the form given in :func:`contribute_tau` then
    To function in Taurex3, it only requires the concrete implementation of:

    - :func:`prepare_each`

    Different forms may require reimplementing
    :func:`contribute` as well as :func:`prepare`

    """

    def __init__(self, name: t.Optional[str] = None):
        """Initialize contribution.

        Parameters
        ----------
        name: str, optional
            Name of contribution for logging.

        """
        name = name or self.__class__.__name__
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self._name = name
        self._total_contribution = None
        self._enabled = True
        self.sigma_xsec = None

    @property
    def order(self) -> int:
        """Computational order.

        Lower numbers are given
        higher priority and are computed first.

        Returns
        -------
        int:
            Order of computation

        """
        return 5

    @property
    def name(self) -> str:
        """Name of the contribution. Identifier for plots"""
        return self._name

    def contribute(
        self,
        model: ForwardModel,
        start_layer: int,
        end_layer: int,
        density_offset: int,
        layer: int,
        density: npt.NDArray[np.float64],
        tau: npt.NDArray[np.float64],
        path_length: t.Optional[npt.NDArray[np.float64]] = None,
    ):
        """Computes an integral for a single layer for the optical depth.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            A forward model

        start_layer: int
            Lowest layer limit for integration

        end_layer: int
            Upper layer limit of integration

        density_offset: int
            offset in density layer

        layer: int
            atmospheric layer being computed

        density: :obj:`array`
            density profile of atmosphere

        tau: :obj:`array`
            optical depth to store result

        path_length: :obj:`array`
            integration length

        """
        self.debug("SIGMA %s", self.sigma_xsec.shape)
        self.debug(
            " %s %s %s %s %s %s %s",
            start_layer,
            end_layer,
            density_offset,
            layer,
            density,
            tau,
            self._ngrid,
        )
        contribute_tau(
            start_layer,
            end_layer,
            density_offset,
            self.sigma_xsec,
            density,
            path_length,
            self._nlayers,
            self._ngrid,
            layer,
            tau,
        )
        self.debug("DONE")

    def build(self, model: ForwardModel) -> None:
        """Called during forward model build phase.

        Does nothing by default

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        """
        pass

    def prepare_each(
        self, model: ForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """
        **Requires implementation**

        Used to prepare each component of the contribution.
        For context when the main ``taurex`` program is run
        with the option each spectra is the component for the
        contribution. For cross-section based contributions,
        the components are each molecule
        Should yield the name of the component and the component itself

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Name of component and component itself

        """

        raise NotImplementedError

    def prepare(self, model: ForwardModel, wngrid: npt.NDArray[np.float64]) -> None:
        """Used to prepare the contribution for the calculation.
        Called before the forward model performs the main optical depth
        calculation. Default behaviour is to loop through :func:`prepare_each`
        and sum all results into a single cross-section.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid
        """

        self._ngrid = wngrid.shape[0]
        self._nlayers = model.nLayers

        sigma_xsec = np.zeros(shape=(self._nlayers, self._ngrid))

        for gas, sigma in self.prepare_each(model, wngrid):
            self.debug("Gas %s", gas)
            self.debug("Sigma %s", sigma)
            sigma_xsec += sigma

        self.sigma_xsec = sigma_xsec
        self.debug("Final sigma is %s", self.sigma_xsec)
        self.info("Done")

    def finalize(self, model: ForwardModel, tau: npt.NDArray[np.float64]):
        """
        Called in the last phase of the calculation, after the optical
        depth has be completely computed.
        """
        pass

    @property
    def sigma(self) -> npt.NDArray[np.float64]:
        """(Effective) Cross-section for contribution."""
        return self.sigma_xsec

    def write(self, output: OutputGroup) -> OutputGroup:
        """Writes contribution class and arguments to file.

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`
            Output object to write to.
        """
        contrib = output.create_group(self.__class__.__name__)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.List[str]:
        """List of input keywords for identification."""
        raise NotImplementedError



"""Mie scattering using Lee et al. 2013 formalism."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.model import OneDForwardModel
from taurex.output import OutputGroup

from .contribution import Contribution


class LeeMieContribution(Contribution):
    """Computes Mie scattering contribution to optica depth.

    Formalism taken from: Lee et al. 2013, ApJ, 778, 97

    Parameters
    ----------

    lee_mie_radius: float
        Particle radius in um

    lee_mie_q: float
        Extinction coefficient

    lee_mie_mix_ratio: float
        Mixing ratio in atmosphere

    lee_mie_bottomP: float
        Bottom of cloud deck in Pa

    lee_mie_topP: float
        Top of cloud deck in Pa


    """

    def __init__(
        self,
        lee_mie_radius: t.Optional[float] = 0.01,
        lee_mie_q: t.Optional[float] = 40,
        lee_mie_mix_ratio: t.Optional[float] = 1e-10,
        lee_mie_bottomP: t.Optional[float] = -1,  # noqa: N803
        lee_mie_topP: t.Optional[float] = -1,
    ) -> None:
        super().__init__("Mie")

        self._mie_radius = lee_mie_radius
        self._mie_q = lee_mie_q
        self._mie_mix = lee_mie_mix_ratio
        self._mie_bottom_pressure = lee_mie_bottomP
        self._mie_top_pressure = lee_mie_topP

    @fitparam(
        param_name="lee_mie_radius",
        param_latex=r"$R^{lee}_{\mathrm{mie}}$",
        default_fit=False,
        default_bounds=[0.01, 0.5],
    )
    def mieRadius(self) -> float:  # noqa: N802
        """Particle radius in um."""
        return self._mie_radius

    @mieRadius.setter
    def mieRadius(self, value: float) -> None:  # noqa: N802
        """Particle radius in um."""
        self._mie_radius = value

    @fitparam(
        param_name="lee_mie_q",
        param_latex=r"$Q_\mathrm{ext}$",
        default_fit=False,
        default_bounds=[-10, 1],
    )
    def mieQ(self) -> float:  # noqa: N802
        """Extinction coefficient."""
        return self._mie_q

    @mieQ.setter
    def mieQ(self, value: float) -> None:  # noqa: N802
        self._mie_q = value

    @fitparam(
        param_name="lee_mie_topP",
        param_latex=r"$P^{lee}_\mathrm{top}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[-1, 1],
    )
    def mieTopPressure(self) -> float:  # noqa: N802
        """Pressure at top of cloud deck in Pa."""
        return self._mie_top_pressure

    @mieTopPressure.setter
    def mieTopPressure(self, value: float) -> None:  # noqa: N802
        """Pressure at top of cloud deck in Pa."""
        self._mie_top_pressure = value

    @fitparam(
        param_name="lee_mie_bottomP",
        param_latex=r"$P^{lee}_\mathrm{bottom}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[-1, 1],
    )
    def mieBottomPressure(self) -> float:  # noqa: N802
        """Pressure at bottom of cloud deck in Pa."""
        return self._mie_bottom_pressure

    @mieBottomPressure.setter
    def mieBottomPressure(self, value: float) -> None:  # noqa: N802
        """Pressure at bottom of cloud deck in Pa."""
        self._mie_bottom_pressure = value

    @fitparam(
        param_name="lee_mie_mix_ratio",
        param_latex=r"$\chi^{lee}_\mathrm{mie}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[-1, 1],
    )
    def mieMixing(self) -> float:  # noqa: N802
        """Mixing ratio in atmosphere."""
        return self._mie_mix

    @mieMixing.setter
    def mieMixing(self, value: float) -> None:  # noqa: N802
        """Mixing ratio in atmosphere."""
        self._mie_mix = value

    def prepare_each(
        self, model: OneDForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Compute and weights the mie opacity for the pressure regions given.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Lee`` and the weighted mie opacity.

        """
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]

        pressure_profile = model.pressureProfile

        bottom_pressure = self.mieBottomPressure
        if bottom_pressure < 0:
            bottom_pressure = pressure_profile[0]

        top_pressure = self.mieTopPressure
        if top_pressure < 0:
            top_pressure = pressure_profile[-1]

        wltmp = 10000 / wngrid

        a = self.mieRadius

        x = 2.0 * np.pi * a / wltmp
        self.debug("wngrid %s", wngrid)
        self.debug("x %s", x)
        q_ext = 5.0 / (self.mieQ * x ** (-4.0) + x ** (0.2))

        sigma_xsec = np.zeros(shape=(self._nlayers, wngrid.shape[0]))

        # This must transform um to the xsec format in TauREx (m2)
        am = a * 1e-6

        sigma_mie = q_ext * np.pi * (am**2.0)

        self.debug("q_ext %s", q_ext)
        self.debug("radius um %s", a)
        self.debug("sigma %s", sigma_mie)

        self.debug("bottome_pressure %s", bottom_pressure)
        self.debug("top_pressure %s", top_pressure)

        cloud_filter = (pressure_profile <= bottom_pressure) & (
            pressure_profile >= top_pressure
        )

        sigma_xsec[cloud_filter, ...] = sigma_mie * self.mieMixing

        self.sigma_xsec = sigma_xsec

        self.debug("final xsec %s", self.sigma_xsec)

        yield "Lee", sigma_xsec

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write output group."""
        contrib = super().write(output)
        contrib.write_scalar("lee_mie_radius", self._mie_radius)
        contrib.write_scalar("lee_mie_q", self._mie_q)
        contrib.write_scalar("lee_mie_mix_ratio", self._mie_mix)
        contrib.write_scalar("lee_mie_bottomP", self._mie_bottom_pressure)
        contrib.write_scalar("lee_mie_topP", self._mie_top_pressure)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Input keywords.""" ""
        return ("LeeMie",)

    BIBTEX_ENTRIES = [
        """
        @article{Lee_2013,
            doi = {10.1088/0004-637x/778/2/97},
            url = {https://doi.org/10.1088%2F0004-637x%2F778%2F2%2F97},
            year = 2013,
            month = {nov},
            publisher = {{IOP} Publishing},
            volume = {778},
            number = {2},
            pages = {97},
            author = {Jae-Min Lee and Kevin Heng and Patrick G. J. Irwin},
            journal = {The Astrophysical Journal},
        """,
    ]



import typing as t

import numpy as np
import numpy.typing as npt

from taurex.types import PathLike
from taurex.util import wnwidth_to_wlwidth

from .array import ArraySpectrum


class TaurexSpectrum(ArraySpectrum):
    """Observation is a taurex spectrum from a HDF5 file.

    An instrument function must have been used for this to work

    """

    def __init__(self, filename: t.Optional[PathLike] = None):
        """Initialize and load from HDF5 file.

        Parameters
        ----------
        filename:
            Filename of HDF5 file

        Raises
        ------
        KeyError
            If instrument output not found in HDF5 file

        """
        super().__init__(self._load_from_hdf5(filename))

    def _load_from_hdf5(self, filename: PathLike) -> npt.NDArray[np.float64]:
        """Load from HDF5 file.

        Parameters
        ----------
        filename:
            Filename of HDF5 file

        Returns
        -------
        np.ndarray[np.float64]
            Array of shape (nbins, 4) with columns:
            1. wavelength (um)
            2. spectral data
            3. error
            4. bin width (um)

        Raises
        ------
        KeyError
            If instrument output not found in HDF5 file

        """
        import h5py

        with h5py.File(filename, "r") as f:
            try:
                wngrid = f["Output"]["Spectra"]["instrument_wngrid"][:]
            except KeyError as e:
                self.error(
                    "Could not find instrument outputs in HDF5, "
                    "this was caused either by the HDF5 being a "
                    "retrieval output or not running with some "
                    "form of instrument in the forward model"
                    " input par file"
                )
                raise KeyError("Instrument output not found") from e

            spectrum = f["Output"]["Spectra"]["instrument_spectrum"][:]
            noise = f["Output"]["Spectra"]["instrument_noise"][:]
            wnwidth = f["Output"]["Spectra"]["instrument_wnwidth"][:]

        wlgrid = 10000 / wngrid

        wlwidth = wnwidth_to_wlwidth(wngrid, wnwidth)

        return np.vstack((wlgrid, spectrum, noise, wlwidth)).T



"""Class for forward spectral modeling."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.binning import Binner
from taurex.core import Citable, DerivedType, Fittable, FittingType
from taurex.log import Logger
from taurex.output import OutputGroup, Writeable
from taurex.types import ModelOutputType

if t.TYPE_CHECKING:
    from taurex.contributions import Contribution
else:
    Contribution = object


class ForwardModel(Logger, Fittable, Writeable, Citable):
    """A base class for producing forward models."""

    def __init__(self, name: str) -> None:
        """Initialise forward model.


        Parameters
        ----------
        name : str
            Name of forward model.

        """
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self.opacity_dict = {}
        self.cia_dict = {}

        self._native_grid = None

        self._derived_parameters = self.derived_parameters()
        self._fitting_parameters = self.fitting_parameters()

        self.contribution_list: t.List[Contribution] = []

    def __getitem__(self, key: str) -> float:
        """Get a fitting parameter value."""
        return self._fitting_parameters[key][2]()

    def __setitem__(self, key: str, value: float) -> None:
        """Set a fitting parameter value."""
        return self._fitting_parameters[key][3](value)

    def defaultBinner(self) -> Binner:  # noqa: N802
        from taurex.binning import NativeBinner

        return NativeBinner()

    def add_contribution(self, contrib: Contribution) -> None:
        """Add a contribution to the forward model."""
        if not isinstance(contrib, Contribution):
            raise TypeError("Is not a a contribution type")
        else:
            if contrib not in self.contribution_list:
                self.contribution_list.append(contrib)
            else:
                raise Exception("Contribution already exists")

    def build(self) -> None:
        """Build forward model."""
        raise NotImplementedError

    def initialize_profiles(self) -> None:
        """Initialize profiles."""
        raise NotImplementedError

    def model(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> ModelOutputType:
        """Computes the forward model for a wngrid"""
        raise NotImplementedError

    def model_contrib(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        t.Dict[
            str,
            t.Tuple[
                npt.NDArray[np.float64], npt.NDArray[np.float64], t.Optional[t.Any]
            ],
        ],
    ]:
        """Computes the forward model for a wngrid with each contribution."""
        raise NotImplementedError

    def model_full_contrib(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        t.Dict[
            str,
            t.List[
                t.Tuple[
                    str,
                    npt.NDArray[np.float64],
                    npt.NDArray[np.float64],
                    t.Optional[t.Any],
                ]
            ],
        ],
    ]:
        """Computes the forward model for a wngrid for each contribution

        Considers each contribution has subcomponents as well.

        """
        raise NotImplementedError

    @property
    def fittingParameters(self) -> t.Dict[str, FittingType]:  # noqa: N802
        """Returns a dictionary of fitting parameters"""
        return self._fitting_parameters

    @property
    def derivedParameters(self) -> t.Dict[str, DerivedType]:  # noqa: N802
        """Returns a dictionary of derived parameters"""
        return self._derived_parameters

    def compute_error(
        self,
        samples: t.Callable[[], float],
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        binner: t.Optional[Binner] = None,
    ) -> t.Tuple[
        t.Dict[str, npt.NDArray[np.float64]], t.Dict[str, npt.NDArray[np.float64]]
    ]:
        """Error of the model and its components given a sample function."""
        return {}, {}

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write forward model to output group."""
        model = output.create_group("ModelParameters")
        model.write_string("model_type", self.__class__.__name__)
        contrib = model.create_group("Contributions")
        for c in self.contribution_list:
            c.write(contrib)

        return model

    def generate_profiles(self) -> t.Dict[str, npt.NDArray[np.float64]]:
        """Generate profiles to store.

        Must return a dictionary of profiles you want to
        store after modeling
        """
        from taurex.util.output import generate_profile_dict

        if hasattr(self, "temperatureProfile"):
            return generate_profile_dict(
                self
            )  # To ensure this change does not break anything
        else:
            return {}

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for forward model."""
        raise NotImplementedError

    def citations(self) -> t.List[str]:
        """Citations for forward model.

        Will either return a list of string or a list of
        :class:`~taurex.data.citation.Citation` objects if
        ``pybtex`` is installed.


        """
        from taurex.core import unique_citations_only

        model_citations = super().citations()
        for c in self.contribution_list:
            model_citations.extend(c.citations())

        return unique_citations_only(model_citations)



import typing as t

import numpy as np
import numpy.typing as npt

from taurex import OutputSize
from taurex.types import ModelOutputType

from .binner import BinDownType, Binner


class LightcurveBinner(Binner):
    """A special class of binning used for lightcurves.

    This is essentially the same as
    :class:`~taurex.binning.nativebinner.NativeBinner`
    but for lightcurve forward models.


    """

    def bindown(
        self,
        wngrid: npt.NDArray[np.float64],
        spectrum: npt.NDArray[np.float64],
        grid_width: t.Optional[npt.NDArray[np.float64]] = None,
        error: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> BinDownType:
        """Does nothing, only returns function arguments"""
        return wngrid, spectrum, error, grid_width

    def generate_spectrum_output(
        self,
        model_output: ModelOutputType,
        output_size: t.Optional[OutputSize] = OutputSize.heavy,
    ):
        """Generate spectrum output for lightcurves.

        Accepts only a lightcurve forward model. Stores the lightcurve
        as well as the spectrum.

        Parameters
        ----------
        model_output: obj:`tuple`
            Result from running a TauREx3 lightcurve forward model

        output_size: :class:`~taurex.taurexdefs.OutputSize`
            Size of the output.


        Returns
        -------
        :obj:`dict`:
            Dictionary of spectra containing both lightcurves
            and spectra.


        """

        output = {}

        wngrid, lightcurve, tau, extra = model_output
        native_grid, native, binned, extra = extra

        output["native_wngrid"] = native_grid
        output["native_wlgrid"] = 10_000 / native_grid
        output["binned_wngrid"] = wngrid
        output["binned_wlgrid"] = 10_000 / wngrid
        output["lightcurve"] = lightcurve
        output["native_spectrum"] = native
        output["binned_spectrum"] = binned
        if output_size > OutputSize.lighter:
            output["binned_tau"] = self.bindown(wngrid, tau)[1]
            if output_size > OutputSize.light:
                output["native_tau"] = tau

        return output



"""Base instrument model class."""
import typing as t

from taurex.binning import BinDownType
from taurex.data.citation import Citable
from taurex.log import Logger
from taurex.model import ForwardModel
from taurex.types import ModelOutputType


class Instrument(Logger, Citable):
    """Instrument noise model.


    *Abstract class*

    Defines some method that transforms
    a spectrum and generates noise.

    """

    def __init__(self) -> None:
        """Constructor."""
        super().__init__(self.__class__.__name__)

    def model_noise(
        self,
        model: ForwardModel,
        model_res: t.Optional[ModelOutputType] = None,
        num_observations: t.Optional[int] = 1,
    ) -> BinDownType:
        """Model noise for a given forward model.

        **Requires implementation**

        For a given forward model (and optional result)
        Resample the spectrum and compute noise profile.

        Parameters
        ----------

        model: :class:`~taurex.model.model.ForwardModel`
            Forward model to pass.

        model_res: :obj:`tuple`, optional
            Result from :func:`~taurex.model.model.ForwardModel.model`

        num_observations: int, optional
            Number of observations to simulate
        """

        raise NotImplementedError

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for instrument."""
        raise NotImplementedError



"""Module for temperature profiles."""

from .file import TemperatureFile
from .guillot import Guillot2010
from .isothermal import Isothermal
from .npoint import NPoint
from .rodgers import Rodgers2000
from .tprofile import TemperatureProfile

__all__ = [
    "Isothermal",
    "Guillot2010",
    "TemperatureProfile",
    "NPoint",
    "Rodgers2000",
    "TemperatureFile",
]



"""
Atmospheric chemistry related modules
"""

from .taurexchemistry import TaurexChemistry
from .gas.constantgas import ConstantGas
from .gas.twolayergas import TwoLayerGas
from .gas.powergas import PowerGas
from .filechemistry import ChemistryFile
from .autochemistry import AutoChemistry


"""Guillot 2010 temperature profile."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.exceptions import InvalidModelException
from taurex.output import OutputGroup

from .tprofile import TemperatureProfile


class Guillot2010(TemperatureProfile):
    """TP profile from Guillot 2010, A&A, 520, A27 (equation 49).

    Using modified 2stream approx.
    from Line et al. 2012, ApJ, 749,93 (equation 19)

    """

    def __init__(
        self,
        T_irr: t.Optional[float] = 1500,  # noqa: N803
        kappa_irr: t.Optional[float] = 0.01,
        kappa_v1: t.Optional[float] = 0.005,
        kappa_v2: t.Optional[float] = 0.005,
        alpha: t.Optional[float] = 0.5,
        T_int: t.Optional[float] = 100,  # noqa: N803
    ):
        """Initialize guillot temperature profile.

        Parameters
        -----------
            T_irr: float
                planet equilibrium temperature
                (Line fixes this but we keep as free parameter)
            kappa_ir: float
                mean infra-red opacity
            kappa_v1: float
                mean optical opacity one
            kappa_v2: float
                mean optical opacity two
            alpha: float
                ratio between kappa_v1 and kappa_v2 downwards radiation stream
            T_int: float
                Internal heating parameter (K)

        """
        super().__init__("Guillot")

        self.T_irr = T_irr

        self.kappa_ir = kappa_irr
        self.kappa_v1 = kappa_v1
        self.kappa_v2 = kappa_v2
        self.alpha = alpha
        self.T_int = T_int

        self._check_values()

    @fitparam(
        param_name="T_irr",
        param_latex=r"$T_\\mathrm{irr}$",
        default_fit=False,
        default_bounds=[1300, 2500],
    )
    def equilTemperature(self) -> float:  # noqa: N802
        """Planet equilibrium temperature"""
        return self.T_irr

    @equilTemperature.setter
    def equilTemperature(self, value: float) -> None:  # noqa: N802
        self.T_irr = value

    @fitparam(
        param_name="kappa_irr",
        param_latex=r"$k_\\mathrm{irr}$",
        default_fit=False,
        default_bounds=[1e-10, 1],
        default_mode="log",
    )
    def meanInfraOpacity(self) -> float:  # noqa: N802
        """mean infra-red opacity"""
        return self.kappa_ir

    @meanInfraOpacity.setter
    def meanInfraOpacity(self, value: float) -> None:  # noqa: N802
        self.kappa_ir = value

    @fitparam(
        param_name="kappa_v1",
        param_latex=r"$k_\\mathrm{1}$",
        default_fit=False,
        default_bounds=[1e-10, 1],
        default_mode="log",
    )
    def meanOpticalOpacity1(self) -> float:  # noqa: N802
        """mean optical opacity one"""
        return self.kappa_v1

    @meanOpticalOpacity1.setter
    def meanOpticalOpacity1(self, value: float) -> None:  # noqa: N802
        """mean optical opacity one"""
        self.kappa_v1 = value

    @fitparam(
        param_name="kappa_v2",
        param_latex=r"$k_\\mathrm{2}$",
        default_fit=False,
        default_bounds=[1e-10, 1],
        default_mode="log",
    )
    def meanOpticalOpacity2(self) -> float:  # noqa: N802
        """mean optical opacity two"""
        return self.kappa_v2

    @meanOpticalOpacity2.setter
    def meanOpticalOpacity2(self, value: float) -> None:  # noqa: N802
        """mean optical opacity two"""
        self.kappa_v2 = value

    @fitparam(
        param_name="alpha",
        param_latex=r"$\\alpha$",
        default_fit=False,
        default_bounds=[0.0, 1.0],
    )
    def opticalRatio(self) -> float:  # noqa: N802
        """ratio between kappa_v1 and kappa_v2."""
        return self.alpha

    @opticalRatio.setter
    def opticalRatio(self, value: float) -> None:  # noqa: N802
        self.alpha = value

    @fitparam(
        param_name="T_int_guillot",
        param_latex="$T^{g}_{int}$",
        default_fit=False,
        default_bounds=[0.0, 1.0],
    )
    def internalTemperature(self) -> float:  # noqa: N802
        """ratio between kappa_v1 and kappa_v2"""
        return self.T_int

    @internalTemperature.setter
    def internalTemperature(self, value: float) -> None:  # noqa: N802
        """ratio between kappa_v1 and kappa_v2"""
        self.T_int = value

    def _check_values(self) -> None:
        """Ensures kappa values are valid.

        Raises
        ------
        InvalidModelException:
            If any kappa is zero

        """
        if self.kappa_ir == 0.0:
            self.warning("Kappa ir is zero")
            raise InvalidModelException("kappa_ir is zero")

        gamma_1 = self.kappa_v1 / (self.kappa_ir)
        gamma_2 = self.kappa_v2 / (self.kappa_ir)

        if gamma_1 == 0.0 or gamma_2 == 0.0:
            self.warning(
                "Gamma is zero. kappa_v1 = %s kappa_v2 = %s" " kappa_ir = %s",
                self.kappa_v1,
                self.kappa_v2,
                self.kappa_ir,
            )
            raise InvalidModelException("Kappa v1/v2/ir values result in zero gamma")

        if self.T_irr < 0 or self.T_int < 0:
            self.warning(
                "Negative temperature input T_irr=%s T_int=%s", self.T_irr, self.T_int
            )
            raise InvalidModelException("Negative temperature input")

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Returns a guillot temperature temperature profile.

        Returns
        --------
        temperature_profile :
            Temperature profile at each layer in Kelvin.

        """

        planet_grav = self.planet.gravity

        self._check_values()

        gamma_1 = self.kappa_v1 / (self.kappa_ir)
        gamma_2 = self.kappa_v2 / (self.kappa_ir)
        tau = self.kappa_ir * self.pressure_profile / planet_grav

        t_int = self.T_int  # todo internal heat parameter looks suspicious..

        def eta(gamma, tau):
            import scipy.special as spe

            part1 = 2.0 / 3.0 + 2.0 / (3.0 * gamma) * (
                1.0 + (gamma * tau / 2.0 - 1.0) * np.exp(-1.0 * gamma * tau)
            )

            part2 = (
                2.0 * gamma / 3.0 * (1.0 - tau**2 / 2.0) * spe.expn(2, (gamma * tau))
            )

            return part1 + part2

        t4 = (
            3.0 * t_int**4 / 4.0 * (2.0 / 3.0 + tau)
            + 3.0 * self.T_irr**4 / 4.0 * (1.0 - self.alpha) * eta(gamma_1, tau)
            + 3.0 * self.T_irr**4 / 4.0 * self.alpha * eta(gamma_2, tau)
        )

        return t4**0.25

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write temperature profile to output."""
        temperature = super().write(output)
        temperature.write_scalar("T_irr", self.T_irr)
        temperature.write_scalar("kappa_irr", self.kappa_ir)
        temperature.write_scalar("kappa_v1", self.kappa_v1)
        temperature.write_scalar("kappa_v2", self.kappa_v2)
        temperature.write_scalar("alpha", self.alpha)
        return temperature

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        return (
            "guillot",
            "guillot2010",
        )

    BIBTEX_ENTRIES = [
        r"""
        @article{guillot,
        author = {{Guillot, T.}},
        title = {On the radiative equilibrium of irradiated planetary atmospheres},
        DOI= "10.1051/0004-6361/200913396",
        url= "https://doi.org/10.1051/0004-6361/200913396",
        journal = {A\&A},
        year = 2010,
        volume = 520,
        pages = "A27",
        month = "",
        }
        """
    ]



"""Base class for optimization/retrieval."""

import math
import typing as t
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from taurex import OutputSize
from taurex.cache import GlobalCache
from taurex.core import DerivedType, FittingType
from taurex.core.priors import LogUniform, Prior, PriorMode, Uniform
from taurex.data.citation import Citable
from taurex.log import Logger, disableLogging, enableLogging
from taurex.model import ForwardModel
from taurex.output import OutputGroup
from taurex.spectrum import BaseSpectrum
from taurex.types import AnyValType

SQRTPI = np.sqrt(2 * np.pi)


class FitParamOutput(t.TypedDict):
    """Dictionary for storing fit parameter output."""

    mean: float
    sigma: float
    value: float
    sigma_m: float
    sigma_p: float
    trace: float
    map: float


@dataclass
class FitParam:
    """Holds information about a fitting parameter."""

    name: str
    latex: str
    fget: t.Callable[[], float]
    fset: t.Callable[[float], None]
    mode: t.Literal["linear", "log"]
    to_fit: bool
    bounds: t.Tuple[float, float]
    prior: Prior = None

    @property
    def fit_prior(self):
        """Get prior for fitting."""
        if self.prior is not None:
            return self.prior
        if self.mode == "linear":
            return Uniform(bounds=self.bounds)
        return LogUniform(lin_bounds=self.bounds)

    @property
    def value(self) -> float:
        """Get value of parameter."""
        return self.fget()

    @value.setter
    def value(self, value: float) -> None:
        """Set value of parameter.

        Parameters
        ----------
        value : float
            Value to set parameter to

        """
        self.fset(self.fit_prior.prior(value))

    @property
    def fit_name(self) -> str:
        """Name for fitting."""
        return (
            self.name
            if self.fit_prior.priorMode is PriorMode.LINEAR
            else f"log_{self.name}"
        )

    @property
    def fit_latex(self) -> str:
        """Latex for fitting."""
        return (
            self.latex
            if self.fit_prior.priorMode is PriorMode.LINEAR
            else f"log({self.latex})"
        )

    @property
    def fit_value(self) -> float:
        """Get value of parameter considering its mode."""
        import math

        return (
            self.value
            if self.fit_prior.priorMode is PriorMode.LINEAR
            else math.log10(self.value)
        )

    @property
    def boundaries(self):
        """Get boundaries of parameter."""
        return self.fit_prior.boundaries()


@dataclass
class DerivedParam:
    """Holds information about a derived parameter."""

    name: str
    latex: str
    fget: t.Callable[[], float]
    compute: bool


def compile_params(
    fitparams: t.Dict[str, FittingType],
    driveparams: t.Dict[str, DerivedType],
    fit_priors: t.Dict[str, Prior] = None,
):
    """Compile fitting and derived parameters."""
    fitting_parameters = [
        FitParam(*params, prior=fit_priors.get(params[0]))
        for params in fitparams.values()
    ]
    derived_parameters = [DerivedParam(*params) for params in driveparams.values()]

    return fitting_parameters, derived_parameters


class Optimizer(Logger, Citable):
    """A base class that handles fitting and optimization of forward models.

    The class handles the compiling and management of fitting parameters in
    forward models, in its current form it cannot fit and requires a class
    derived from it to implement the :func:`compute_fit` function.

    """

    def __init__(
        self,
        name: str,
        observed: t.Optional[BaseSpectrum] = None,
        model: t.Optional[ForwardModel] = None,
        sigma_fraction: t.Optional[float] = 0.1,
    ) -> None:
        """Initilize optimizer.

        Parameters
        ----------
        name:
            Name of optimizer for logging

        observed:
            Observed spectrum

        model:
            Model to be optimized

        sigma_fraction:
            Fraction of weights to use in computing the error. (Default: 10%)


        """
        super().__init__(name)

        self._model_callback = None
        self._sigma_fraction = sigma_fraction
        self._fit_priors = {}
        self.avail_fit_parameters: t.Dict[str, FitParam] = {}
        self.avail_derived_parameters: t.Dict[str, DerivedParam] = {}
        self._model: ForwardModel = None
        self._observed: BaseSpectrum = None
        self.set_model(model)
        self.set_observed(observed)

    def set_model(self, model: ForwardModel) -> None:
        """Sets the model to be optimized/fit

        Parameters
        ----------
        model:
            The forward model we wish to optimize

        """
        self._model: ForwardModel = model
        self._compile_params()

    def set_observed(self, observed: BaseSpectrum) -> None:
        """Sets the observation to optimize the model to.

        Parameters
        ----------
        observed:
            Observed spectrum we will optimize to

        """
        self._observed: BaseSpectrum = observed
        if observed is not None:
            self._binner = observed.create_binner()
        self._compile_params()

    def compile_params(self) -> None:
        """Dummy, does nothing and will be depcreated"""
        import warnings

        warnings.warn(
            "compile_params is deprecated and will be removed in future versions. ",
            DeprecationWarning,
            stacklevel=2,
        )

    def _compile_params(self) -> None:
        """Compile parameters for fitting from model and observation."""
        self.info("Initializing parameters")
        model_fit, model_derive = [], []
        if self._model is not None:
            (
                model_fit,
                model_derive,
            ) = compile_params(
                self._model.fittingParameters,
                self._model.derivedParameters,
                self._fit_priors,
            )

        obs_fit, obs_deriv = [], []
        if self._observed is not None:
            obs_fit, obs_deriv = compile_params(
                self._observed.fittingParameters,
                self._observed.derivedParameters,
                self._fit_priors,
            )

        fitting_params = model_fit + obs_fit
        derived_params = model_derive + obs_deriv
        # self.fitting_priors.extend(obs_prior)
        self.avail_fit_parameters = {p.name: p for p in fitting_params}
        self.avail_derived_parameters = {p.name: p for p in derived_params}

        self.info("-------FITTING---------------")
        self.info("Fitting Parameters available:")
        for params in self.avail_fit_parameters.values():
            self.info(
                "%s, Value: %s Type: %s, Params: %s",
                params.name,
                params.fget(),
                params.fit_prior.__class__.__name__,
                params.fit_prior.params(),
            )
        self.info("-------DERIVED---------------")
        self.info("Derived Parameters available:")
        for params in self.avail_derived_parameters.values():
            self.info("%s, Value: %s", params.name, params.fget())

    @property
    def fitting_parameters(self) -> t.List[FitParam]:
        """Returns the list of fitting parameters."""
        return [f for f in self.avail_fit_parameters.values() if f.to_fit]

    @property
    def derived_parameters(self) -> t.List[DerivedParam]:
        """Returns the list of derived parameters."""
        return [f for f in self.avail_derived_parameters.values() if f.compute]

    @property
    def fitting_priors(self) -> t.List[Prior]:
        """Returns the list of fitting priors.

        This is mostly for compatibility with the old code.

        """
        return [f.fit_prior for f in self.fitting_parameters]

    def prior_transform(self, cube: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """Transforms a unit cube to a prior cube.

        Employ mixins to override this method.

        """
        return np.array(
            [fit.fit_prior.sample(c) for fit, c in zip(self.fitting_parameters, cube)]
        )

    def log_likelihood(self, parameters: npt.ArrayLike) -> float:
        r"""Log likelihood function.

        Computed as:

        .. math::

            \log \mathcal{L} = -\sum_i \sigma_i \sqrt{2\pi}) - \frac{1}{2} \chi^2

        Employ mixins to override this method.

        """
        parameters = np.asarray(parameters)
        data = self._observed.spectrum
        datastd = self._observed.errorBar

        chi_t = self.chisq_trans(parameters, data, datastd)
        loglike = -np.sum(np.log(datastd * SQRTPI)) - 0.5 * chi_t
        return loglike

    def update_model(self, fit_params: t.List[float]) -> None:
        """Updates the model with new parameters

        Parameters
        ----------
        fit_params : :obj:`list`
            A list of new values to apply to the model. The list of values are
            assumed to be in the same order as the parameters given by
            :func:`fit_names`


        """
        if len(fit_params) != len(self.fitting_parameters):
            self.error(
                "Trying to update model with more fitting parameters" " than enabled"
            )
            self.error(
                f"No. enabled parameters:{len(self.fitting_parameters)}"
                f" Update length: {len(fit_params)}"
            )
            raise ValueError(
                "Trying to update model with more fitting" " parameters than enabled"
            )

        for value, param in zip(fit_params, self.fitting_parameters):
            param.value = value

    @property
    def fit_values_nomode(self) -> t.List[float]:
        """Returns a list of the current values of a fitting parameter.
        Regardless of the ``mode`` setting

        Returns
        -------
        :obj:`list`:
            List of each value of a fitting parameter

        """

        return [c.value for c in self.fitting_parameters]

    @property
    def fit_values(self) -> t.List[float]:
        """Returns a list of the current values of a fitting parameter.

        This respects the ``mode`` setting

        Returns
        -------
        :obj:`list`:
            List of each value of a fitting parameter

        """

        return [c.fit_value for c in self.fitting_parameters]

    @property
    def fit_boundaries(self):
        """

        Returns the fitting boundaries of the parameter

        Returns
        -------
        :obj:`list`:
            List of boundaries for each fitting parameter. It takes the form of
            a python :obj:`tuple` with the form
            ( ``bound_min`` , ``bound_max`` )

        """
        return [
            c[-1] if c[4] == "linear" else (math.log10(c[-1][0]), math.log10(c[-1][1]))
            for c in self.fitting_parameters
        ]

    @property
    def fit_names(self) -> t.List[str]:
        """Returns the names of the model parameters we will be fitting

        Returns
        -------
        :obj:`list`:
            List of names of parameters that will be fit

        """

        return [c.fit_name for c in self.fitting_parameters]

    @property
    def fit_latex(self) -> t.List[str]:
        """Returns the names of the parameters in LaTeX format."""
        return [c.fit_latex for c in self.fitting_parameters]

    @property
    def derived_names(self) -> t.List[str]:
        """Return names for derived parameters."""
        return [c.name for c in self.derived_parameters]

    @property
    def derived_latex(self) -> t.List[str]:
        """Returns a list of the current values of a fitting parameter."""
        return [c.latex for c in self.derived_parameters]

    @property
    def derived_values(self) -> t.List[float]:
        """Returns current values of derived parameters."""
        return [c.fget() for c in self.derived_parameters]

    def enable_fit(self, parameter: str) -> None:
        """Enables fitting of the parameter"""
        self.avail_fit_parameters[parameter].to_fit = True

    def enable_derived(self, parameter: str) -> None:
        """Enables computation of derived parameter."""
        self.avail_derived_parameters[parameter].compute = True

    def disable_fit(self, parameter: str) -> None:
        """Disables fitting of the parameter."""
        self.avail_fit_parameters[parameter].to_fit = False

    def disable_derived(self, parameter: str) -> None:
        """Disables computation of derived parameter."""
        self.avail_derived_parameters[parameter].compute = False

    def set_boundary(self, parameter: str, new_boundaries: t.List[float]) -> None:
        """Sets the boundary of the parameter."""
        self.avail_fit_parameters[parameter].bounds = new_boundaries

    def set_factor_boundary(self, parameter: str, factors: t.List[float]) -> None:
        """Sets the boundary of the parameter based on a factor of value."""
        param = self.avail_fit_parameters[parameter]
        bounds = (param.value * factors[0], param.value * factors[1])
        param.bounds = (min(bounds), max(bounds))

    def set_mode(self, parameter: str, new_mode: t.Literal["linear", "log"]) -> None:
        """Sets the fitting mode of a parameter."""
        self.avail_fit_parameters[parameter].mode = new_mode

    def set_prior(self, parameter: str, prior: Prior) -> None:
        """Sets the prior of a parameter."""
        self.avail_fit_parameters[parameter].prior = prior

    def chisq_trans(
        self,
        fit_params: npt.ArrayLike,
        data: npt.NDArray[np.float64],
        datastd: npt.NDArray[np.float64],
    ) -> float:
        """Computes the Chi-Squared of model and observation

        Computes the Chi-Squared between the forward model and
        observation. The steps taken are:
            1. Forward model (FM) is updated with :func:`update_model`
            2. FM is then computed at its native grid then binned.
            3. Chi-squared between FM and observation is computed
        """
        from taurex.exceptions import InvalidModelException

        self.update_model(fit_params)
        obs_bins = self._observed.wavenumberGrid

        try:
            _, final_model, _, _ = self._binner.bin_model(
                self._model.model(wngrid=obs_bins)
            )
        except InvalidModelException:
            return np.nan

        res = (data.ravel() - final_model.ravel()) / datastd.ravel()

        reject_nan = GlobalCache()["reject_spectral_nan"]

        if reject_nan and np.any(np.isnan(res)):
            return np.nan

        res = np.nansum(res * res)
        if res == 0:
            res = np.nan

        return res

    def compute_fit(self) -> t.Any:
        """Main compute fit function.

        Unimplemented. When inheriting this should be overwritten
        to perform the actual fit.

        Raises
        ------
        NotImplementedError
            Raised when a derived class does override this function

        """
        raise NotImplementedError

    def fit(self, output_size=OutputSize.heavy) -> t.Dict[str, AnyValType]:
        """Performs retrieval."""
        import time

        from tabulate import tabulate

        fit_names = self.fit_names

        prior_type = [p.__class__.__name__ for p in self.fitting_priors]
        args = [p.params() for p in self.fitting_priors]

        fit_values = self.fit_values
        self.info("")
        self.info("-------------------------------------")
        self.info("------Retrieval Parameters-----------")
        self.info("-------------------------------------")
        self.info("")
        self.info("Dimensionality of fit: %s", len(fit_names))
        self.info("")

        output = tabulate(
            zip(fit_names, fit_values, prior_type, args),
            headers=["Param", "Value", "Type", "Args"],
        )

        self.info("\n%s\n\n", output)
        self.info("")
        start_time = time.time()
        disableLogging()
        # Compute fit here
        self.compute_fit()

        enableLogging()
        end_time = time.time()
        self.info("Sampling time %s s", end_time - start_time)
        solution = self.generate_solution(output_size=output_size)
        self.info("")
        self.info("-------------------------------------")
        self.info("------Final results------------------")
        self.info("-------------------------------------")
        self.info("")
        self.info("Dimensionality of fit: %s", len(fit_names))
        self.info("")
        for idx, optimized_map, optimized_median, _ in self.get_solution():
            self.info("\n%s", f"---Solution {idx}------")
            output = tabulate(
                zip(fit_names, optimized_map, optimized_median),
                headers=["Param", "MAP", "Median"],
            )
            self.info("\n%s\n\n", output)
        return solution

    def write_optimizer(self, output: OutputGroup) -> OutputGroup:
        """Writes optimizer settings under the 'Optimizer' heading in an output file.

        Parameters
        ----------
        output:
            Group (or root) in output file to write to
        """
        output.write_string("optimizer", self.__class__.__name__)
        output.write_string_array("fit_parameter_names", self.fit_names)
        output.write_string_array("fit_parameter_latex", self.fit_latex)
        output.write_array(
            "fit_boundary_low",
            np.array([x.boundaries()[0] for x in self.fitting_priors]),
        )
        output.write_array(
            "fit_boundary_high",
            np.array([x.boundaries()[1] for x in self.fitting_priors]),
        )
        if len(self.derived_names) > 0:
            output.write_string_array("derived_parameter_names", self.derived_names)
            output.write_string_array("derived_parameter_latex", self.derived_latex)

        return output

    def write_fit(self, output: OutputGroup) -> OutputGroup:
        """Writes basic fitting parameters into output.

        Parameters
        ----------
        output:
            Group (or root) in output file to write to

        """
        fit = output.create_group("FitParams")
        fit.write_string("fit_format", self.__class__.__name__)
        fit.write_string_array("fit_parameter_names", self.fit_names)
        fit.write_string_array("fit_parameter_latex", self.fit_latex)
        fit.write_array(
            "fit_boundary_low", np.array([x[0] for x in self.fit_boundaries])
        )
        fit.write_array(
            "fit_boundary_high", np.array([x[1] for x in self.fit_boundaries])
        )

        # This is the last sampled value ... should not be recorded to avoid confusion.
        # fit.write_list('fit_parameter_values',self.fit_values)
        # fit.write_list('fit_parameter_values_nomode',self.fit_values_nomode)
        return output

    def generate_profiles(
        self, solution: int, binning: npt.NDArray[np.float64]
    ) -> t.Tuple[
        t.Dict[str, npt.NDArray[np.float64]], t.Dict[str, npt.NDArray[np.float64]]
    ]:
        """Generates sigma plots for profiles"""
        from taurex import mpi

        sample_list: t.List[float] = []

        if mpi.get_rank() == 0:
            sample_list = list(self.sample_parameters(solution))

        sample_list = mpi.broadcast(sample_list)

        self.debug("We all got %s", sample_list)

        self.info("------------Variance generation step------------------")

        self.info("We are sampling %s points for the profiles", len(sample_list))

        rank = mpi.get_rank()
        size = mpi.nprocs()

        enableLogging()

        self.info(
            "I will only iterate through partitioned %s "
            "points (the rest is in parallel)",
            len(sample_list) // size,
        )

        disableLogging()

        def sample_iter():
            count = 0
            for parameters, weight in sample_list[rank::size]:
                self.update_model(parameters)
                enableLogging()
                if rank == 0 and count % 10 == 0 and count > 0:
                    self.info(
                        "Progress %.3f", count * 100.0 / (len(sample_list) / size)
                    )
                disableLogging()
                yield weight
                count += 1

        return self._model.compute_error(
            sample_iter, wngrid=binning, binner=self._binner
        )

    def generate_solution(  # noqa: C901
        self, output_size=OutputSize.heavy
    ) -> t.Dict[str, AnyValType]:
        """Generates a dictionary with all solutions and other useful parameters."""
        from taurex.util.output import store_contributions

        solution_dict = {}

        self.info("Post-processing - Generating spectra and profiles")

        # Loop through each solution, grab optimized parameters and anything
        # else we want to store
        for solution, optimized_map, optimized_median, values in self.get_solution():
            enableLogging()
            self.info("Computing solution %s", solution)
            sol_values = {}

            # Include extra stuff we might want to store (provided by the child)
            for k, v in values:
                sol_values[k] = v

            # Update the model with optimized map values
            self.update_model(optimized_map)

            opt_result = self._model.model(cutoff_grid=False)  # Run the model

            sol_values["Spectra"] = self._binner.generate_spectrum_output(
                opt_result, output_size=output_size
            )

            try:
                sol_values["Spectra"]["Contributions"] = store_contributions(
                    self._binner, self._model, output_size=output_size - 3
                )
            except Exception as e:
                self.warning("Not bothering to store contributions since its broken")
                self.warning("%s ", str(e))

            # Update with the optimized median
            self.update_model(optimized_median)

            self._model.model(cutoff_grid=False)

            # Store profiles here
            sol_values["Profiles"] = self._model.generate_profiles()
            profile_dict, spectrum_dict = self.generate_profiles(
                solution, self._observed.wavenumberGrid
            )

            for k, v in profile_dict.items():
                if k in sol_values["Profiles"]:
                    sol_values["Profiles"][k].update(v)
                else:
                    sol_values["Profiles"][k] = v

            for k, v in spectrum_dict.items():
                if k in sol_values["Spectra"]:
                    sol_values["Spectra"][k].update(v)
                else:
                    sol_values["Spectra"][k] = v

            solution_dict[f"solution{solution}"] = sol_values

        if len(self.derived_names) > 0:
            # solution_dict[f'solution{solution}']['derived_params'] = {}
            # Compute derived
            for (
                solution,
                _,
                _,
                _,
            ) in self.get_solution():
                solution_dict[f"solution{solution}"]["derived_params"] = {}
                derived_dict = self.compute_derived_trace(solution)
                if derived_dict is None:
                    continue
                solution_dict[f"solution{solution}"]["derived_params"].update(
                    derived_dict
                )

        enableLogging()
        self.info("Post-processing - Complete")

        return solution_dict

    def compute_derived_trace(self, solution: int) -> t.Dict[str, AnyValType]:
        """Computes derived parameters from traces."""
        from taurex import mpi
        from taurex.util import quantile_corner

        enableLogging()

        samples = self.get_samples(solution)
        weights = self.get_weights(solution)
        len_samples = len(samples)

        rank = mpi.get_rank()

        num_procs = mpi.nprocs()

        count = 0

        derived_param = {p: ([], []) for p in self.derived_names}

        if len(self.derived_names) == 0:
            return

        self.info("Computing derived parameters......")
        disableLogging()
        for idx in range(rank, len_samples, num_procs):
            enableLogging()
            if rank == 0 and count % 10 == 0 and count > 0:
                self.info(f"Progress {idx * 100.0 / len_samples}%")
            disableLogging()

            parameters = samples[idx]
            weight = weights[idx]
            self.update_model(parameters)
            self._model.initialize_profiles()
            for p, v in zip(self.derived_names, self.derived_values):
                derived_param[p][0].append(v)
                derived_param[p][1].append(weight)

        result_dict = {}

        sorted_weights = weights.argsort()

        for param, (trace, w) in derived_param.items():
            # I cant remember why this works
            all_trace = np.array(mpi.allreduce(trace, op="SUM"))
            # I cant remember why this works
            all_weight = np.array(mpi.allreduce(w, op="SUM"))

            all_weight_sort = all_weight.argsort()

            # Sort them into the right order
            all_weight[sorted_weights] = all_weight[all_weight_sort]
            all_trace[sorted_weights] = all_trace[all_weight_sort]

            q_16, q_50, q_84 = quantile_corner(
                np.array(all_trace), [0.16, 0.5, 0.84], weights=np.array(all_weight)
            )

            mean = np.average(all_trace, weights=all_weight, axis=0)

            derived = {
                "value": q_50,
                "sigma_m": q_50 - q_16,
                "sigma_p": q_84 - q_50,
                "trace": all_trace,
                "mean": mean,
            }
            result_dict[f"{param}_derived"] = derived
        return result_dict

    def sample_parameters(
        self, solution: int
    ) -> t.Generator[t.Tuple[npt.NDArray[np.float64], float], None, None]:
        """Read traces and weights and return a random ``sigma_fraction`` sample

        Parameters
        ----------
        solution:
            a solution output from sampler

        Yields
        ------
        traces: :obj:`array`
            Traces of a particular sample

        weight: float
            Weight of sample

        """
        from taurex.util import random_int_iter

        samples = self.get_samples(solution)
        weights = self.get_weights(solution)

        iterator = random_int_iter(samples.shape[0], self._sigma_fraction)
        for x in iterator:
            w = weights[x] + 1e-300

            yield samples[x, :], w

    def get_solution(
        self,
    ) -> t.Generator[
        t.Tuple[
            int,
            npt.NDArray[np.float64],
            npt.NDArray[np.float64],
            t.Dict[str, AnyValType],
        ],
        None,
        None,
    ]:
        """Generator for solutions.

        ** Requires implementation **

        Generator for solutions and their
        median and MAP values

        Yields
        ------

        solution_no:
            Solution number

        map:
            Map values

        median:
            Median values

        extra:
            List of tuples of extra information to store.
            Must be of form ``(name, data)``



        """
        raise NotImplementedError

    def get_samples(self, solution_id: int) -> npt.NDArray[np.float64]:
        """Get traces for a particular solution."""
        raise NotImplementedError

    def get_weights(self, solution_id: int) -> npt.NDArray[np.float64]:
        """Get weights for a particular solution."""
        raise NotImplementedError

    def write(self, output: OutputGroup) -> None:
        """Creates  'Optimizer' and writes output

        Parameters
        ----------
        output :
            Group (or root) in output file to write to



        """
        opt = output.create_group("Optimizer")
        self.write_optimizer(opt)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for optimizer."""
        raise NotImplementedError



import importlib
import io
import multiprocessing as mp
import os
import sys
import tempfile
import typing as t
from contextlib import contextmanager
from queue import Empty, Full, Queue

from taurex.log import Logger

from .module import getattr_recursive, runfunc_recursive, setattr_recursive


@contextmanager
def stdout_redirector(stream: t.IO):
    # The original fd stdout points to. Usually 1 on POSIX systems.
    try:
        original_stdout_fd = sys.stdout.fileno()
    except io.UnsupportedOperation:
        yield
        return

    def _redirect_stdout(to_fd):
        """Redirect stdout to the given file descriptor."""
        # Flush the C-level buffer stdout
        # libc.fflush(c_stdout)
        # Flush and close sys.stdout - also closes the file descriptor (fd)
        sys.stdout.close()
        # Make original_stdout_fd point to the same file as to_fd
        os.dup2(to_fd, original_stdout_fd)
        # Create a new sys.stdout that points to the redirected fd
        sys.stdout = io.TextIOWrapper(os.fdopen(original_stdout_fd, "wb"))

    # Save a copy of the original stdout fd in saved_stdout_fd
    saved_stdout_fd = os.dup(original_stdout_fd)
    try:
        # Create a temporary file and redirect stdout to it
        tfile = tempfile.TemporaryFile(mode="w+b")
        _redirect_stdout(tfile.fileno())
        # Yield to caller, then redirect stdout back to the saved fd
        yield
        _redirect_stdout(saved_stdout_fd)
        # Copy contents of temporary file to the given stream
        tfile.flush()
        tfile.seek(0, io.SEEK_SET)
        if stream is not None:
            stream.write(tfile.read())
    finally:
        tfile.close()
        os.close(saved_stdout_fd)


class FortranStopException(Exception):  # noqa
    """Raised when the fortran process crashes or is stopped."""

    pass


class StreamToLogQueue:
    """Fake file-like stream object that redirects writes to a queue."""

    def __init__(self, log_queue):
        self.log_queue = log_queue

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            out = line.rstrip()

            if isinstance(out, bytes):
                out = out.decode("utf-8")
            try:
                self.log_queue.put_nowait(out)
            except Full:
                pass


class SafeFortranProcess(mp.Process):
    """A process that can safely call fortran functions.

    This is used to safely call fortran functions from python. It is used
    to overcome the STOP issue with fortran code that kills the interpreter.

    """

    def __init__(
        self,
        work_queue: Queue,
        output_queue: Queue,
        shutdown_event: mp.Event,
        module_string: str,
        log_queue: t.Optional[Queue] = None,
    ):
        super().__init__()
        self._module_string = module_string
        self.work_queue = work_queue
        self.output_queue = output_queue
        self.shutdown_event = shutdown_event
        self.log_queue = log_queue

    def run(self) -> None:  # noqa: C901
        """Run the process."""
        timeout = 1.0

        mod = importlib.import_module(self._module_string)
        if self.log_queue is not None:
            logger = StreamToLogQueue(self.log_queue)

        while not self.shutdown_event.is_set():
            try:
                item = self.work_queue.get(block=True, timeout=timeout)
            except Empty:
                continue

            if item == "END":
                break

            task = item[0]
            attr = item[1]
            out = None, None
            if task == "GET":
                try:
                    out = "Success", getattr_recursive(mod, attr)
                except Exception as e:
                    out = "Exception", e

            elif task == "SET":
                value = item[2]
                try:
                    out = "Success", setattr_recursive(mod, attr, value)
                except Exception as e:
                    out = "Exception", e
            elif task == "CALL":
                args = item[2]
                kwargs = item[3]
                try:
                    with stdout_redirector(logger):
                        out = "Success", runfunc_recursive(mod, attr, *args, **kwargs)
                except Exception as e:
                    out = "Exception", str(e)

            self.output_queue.put(out)


class SafeFortranCaller:
    """A safe fortran caller."""

    STOP_WAIT_SECS = 2.0

    def __init__(self, module: str, logger: t.Optional[Logger] = None) -> None:
        """Create a safe fortran caller.

        Parameters
        ----------
        module : str
            The module to call
        logger : t.Optional[Logger], optional
            A logger to use, by default None


        """
        self._module = module
        self._process = None
        self._message_queue = None
        self._output_queue = None
        self._log_queue = None
        self.shutdown_event = None
        self.logger = logger

    @property
    def process_queues(self):
        """Get the process queues."""
        if self._process is None:
            self.cleanup()
            self._message_queue = mp.Queue()
            self.shutdown_event = mp.Event()
            self._output_queue = mp.Queue()
            self._log_queue = mp.Queue()
            self._process = SafeFortranProcess(
                self._message_queue,
                self._output_queue,
                self.shutdown_event,
                self._module,
                self._log_queue,
            )
            self._process.daemon = True
            self._process.start()
        return self._process, self._message_queue, self._output_queue

    def get_val(self, attr: str) -> t.Any:
        """Get a value from the fortran module."""
        p, m, o = self.process_queues

        success, value = "Crash", None
        timeout = 1.0
        m.put(("GET", attr))

        while p.is_alive():
            try:
                success, value = o.get(block=True, timeout=timeout)
                break
            except Empty:
                continue
        if success == "Crash":
            self.cleanup()
            raise FortranStopException
        elif success == "Exception":
            self.cleanup()
            raise Exception(value)
        elif success == "Success":
            return value
        else:
            self.cleanup()
            raise Exception(f"Unknown exception {attr}-{success}-{value}")

    def set_val(self, attr: str, set_value: t.Any) -> None:
        """Set a value in the fortran module."""
        p, m, o = self.process_queues

        success, value = "Crash", None
        timeout = 1.0
        m.put(("SET", attr, set_value))

        while p.is_alive():
            try:
                success, value = o.get(block=True, timeout=timeout)
                break
            except Empty:
                continue
        if success == "Crash":
            self.cleanup()
            raise FortranStopException
        elif success == "Exception":
            self.cleanup()
            raise Exception(value)
        elif success == "Success":
            return value
        else:
            self.cleanup()
            raise Exception(
                f"Unknown exception {attr}-{set_value}-" f"{success}-{value}"
            )

    def call(self, attr: str, *args, **kwargs) -> t.Any:
        """Call a function in the fortran module."""
        p, m, o = self.process_queues

        success, value = "Crash", None
        timeout = 1.0
        m.put(("CALL", attr, args, kwargs))

        while p.is_alive():
            self.process_logs()
            try:
                success, value = o.get(block=True, timeout=timeout)
                break
            except Empty:
                continue

        self.process_logs()

        if success == "Crash":
            self.cleanup()
            raise FortranStopException
        elif success == "Exception":
            self.cleanup()
            raise Exception(value)
        elif success == "Success":
            return value
        else:
            self.cleanup()
            raise Exception(f"Unknown exception {attr}" f"-{success}-{value}")

    def process_logs(self) -> None:
        """Process logs from the fortran module."""
        try:
            item = self._log_queue.get(block=False)
        except Empty:
            item = None

        while item:
            if self.logger is not None:
                self.logger.info(item)
            try:
                item = self._log_queue.get(block=False)
            except Empty:
                break

    def cleanup_processes(self) -> None:
        """Cleanup the processes."""
        if self.shutdown_event is not None:
            self.shutdown_event.set()

        end_time = self.STOP_WAIT_SECS

        if self._process is not None:
            self._process.join(end_time)
            if self._process.is_alive():
                self._process.terminate()
        self._process = None
        self.shutdown_event = None

    def _cleanup_queue(self, queue: Queue) -> None:
        """Cleanup a queue."""
        try:
            item = queue.get(block=False)
        except Empty:
            item = None

        while item:
            try:
                item = queue.get(block=False)
            except Empty:
                break

        queue.close()
        queue.join_thread()

    def cleanup_queues(self) -> None:
        """Cleanup all queues."""
        if self._message_queue is not None:
            self._cleanup_queue(self._message_queue)
            self._message_queue = None
        if self._output_queue is not None:
            self._cleanup_queue(self._output_queue)
            self._output_queue = None
        if self._log_queue is not None:
            self._cleanup_queue(self._log_queue)
            self._log_queue = None

    def cleanup(self) -> None:
        """Cleanup the fortran caller."""
        self.cleanup_processes()
        try:
            self.cleanup_queues()
        except OSError:
            pass

    def __del__(self):
        """Cleanup on deletion."""
        self.cleanup()



import typing as t

import numpy as np
import numpy.typing as npt

from taurex.util.math import (
    intepr_bilin,
    interp_exp_and_lin,
    interp_exp_only,
    interp_lin_only,
)

from .opacity import Opacity

InterpModeType = t.Literal["linear", "exp"]


class InterpolatingOpacity(Opacity):
    """Provides interpolation methods."""

    def __init__(
        self, name: str, interpolation_mode: t.Optional[InterpModeType] = "linear"
    ) -> None:
        """Initialize interpolating opacity.

        Parameters
        ----------
        name: str
            Name of opacity for logging.
        interpolation_mode: str
            Interpolation mode. Either 'linear' or 'exp'.


        """
        super().__init__(name)
        self._interp_mode = interpolation_mode

    @property
    def pressureMax(self) -> float:  # noqa: N802
        """Maximum pressure in opacity grid."""
        return self.pressureGrid[-1]

    @property
    def pressureMin(self) -> float:  # noqa: N802
        """Minimum pressure in opacity grid."""
        return self.pressureGrid[0]

    @property
    def temperatureMax(self) -> float:  # noqa: N802
        """Maximum temperature in opacity grid."""
        return self.temperatureGrid[-1]

    @property
    def temperatureMin(self) -> float:  # noqa: N802
        """Minimum temperature in opacity grid."""
        return self.temperatureGrid[0]

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Opacity grid."""
        raise NotImplementedError

    @property
    def logPressure(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Log pressure grid."""
        return np.log10(self.pressureGrid)

    @property
    def pressureBounds(self) -> t.Tuple[float, float]:  # noqa: N802
        """Pressure bounds."""
        return self.logPressure.min(), self.logPressure.max()

    @property
    def temperatureBounds(self) -> t.Tuple[float, float]:  # noqa: N802
        """Temperature bounds."""
        return self.temperatureGrid.min(), self.temperatureGrid.max()

    def find_closest_index(
        self, temperature: float, pressure: float
    ) -> t.Tuple[int, int, int, int]:
        """Find closest pair index to temperature and pressure.

        Closest pair index is the index on either side of
        temperature and pressure.


        """
        from taurex.util import find_closest_pair

        # t_min = self.temperatureGrid.searchsorted(T, side='right')-1
        # t_min = max(0, t_min)
        # t_max = t_min+1
        # t_max = min(len(self.temperatureGrid)-1, t_max)
        # p_min = self.pressureGrid.searchsorted(P, side='right')-1
        # p_min = max(0, p_min)
        # p_max = p_min+1
        # p_max = min(len(self.pressureGrid)-1, p_max)

        t_min, t_max = find_closest_pair(self.temperatureGrid, temperature)
        p_min, p_max = find_closest_pair(self.logPressure, pressure)

        return t_min, t_max, p_min, p_max

    def set_interpolation_mode(self, interp_mode: InterpModeType) -> None:
        """Set interpolation mode."""
        self._interp_mode = interp_mode.strip()

    def interp_temp_only(
        self,
        temperature: float,
        temp_idx_min: int,
        temp_idx_max: int,
        pressure_idx: int,
        filt: npt.NDArray[np.bool_],
    ) -> npt.NDArray[np.float64]:
        """Interpolate temperature only."""
        temperature_max = self.temperatureGrid[temp_idx_max]
        temperature_min = self.temperatureGrid[temp_idx_min]
        fx0 = self.xsecGrid[pressure_idx, temp_idx_min, filt]
        fx1 = self.xsecGrid[pressure_idx, temp_idx_max, filt]

        if self._interp_mode == "linear":
            return interp_lin_only(
                fx0, fx1, temperature, temperature_min, temperature_max
            )
        elif self._interp_mode == "exp":
            return interp_exp_only(
                fx0, fx1, temperature, temperature_min, temperature_max
            )
        else:
            raise ValueError(f"Unknown interpolation mode {self._interp_mode}")

    def interp_pressure_only(
        self,
        pressure: float,
        p_idx_min: int,
        p_idx_max: int,
        temperature_idx: int,
        filt: npt.NDArray[np.bool_],
    ) -> npt.NDArray[np.float64]:
        """Interpolate pressure only."""
        pressure_max = self.logPressure[p_idx_max]
        pressure_min = self.logPressure[p_idx_min]
        fx0 = self.xsecGrid[p_idx_min, temperature_idx, filt]
        fx1 = self.xsecGrid[p_idx_max, temperature_idx, filt]

        return interp_lin_only(fx0, fx1, pressure, pressure_min, pressure_max)

    def interp_bilinear_grid(
        self,
        temperature: float,
        pressure: float,
        t_idx_min: int,
        t_idx_max: int,
        p_idx_min: int,
        p_idx_max: int,
        wngrid_filter: npt.NDArray[np.bool_] = None,
    ) -> npt.NDArray[np.float64]:
        """Interpolate both temperature and pressure."""
        self.debug(
            "Interpolating %s %s %s %s %s %s",
            temperature,
            pressure,
            t_idx_min,
            t_idx_max,
            p_idx_min,
            p_idx_max,
        )

        check_pressure_max = pressure >= self.pressureMax
        check_temperature_max = temperature >= self.temperatureMax

        min_pressure, max_pressure = self.pressureBounds
        min_temperature, max_temperature = self.temperatureBounds

        check_pressure_max = pressure >= max_pressure
        check_temperature_max = temperature >= max_temperature

        check_pressure_min = pressure < min_pressure
        check_temperature_min = temperature < min_temperature

        self.debug(
            "Check pressure min/max %s/%s", check_pressure_min, check_pressure_max
        )
        self.debug(
            "Check temperature min/max %s/%s",
            check_temperature_min,
            check_temperature_max,
        )
        # Are we both max?
        if check_pressure_max and check_temperature_max:
            self.debug("Maximum Temperature pressure reached. Using last")
            return self.xsecGrid[-1, -1, wngrid_filter].ravel()

        if check_pressure_min and check_temperature_min:
            return np.zeros_like(self.xsecGrid[0, 0, wngrid_filter]).ravel()

        # Max pressure
        if check_pressure_max:
            self.debug("Max pressure reached. Interpolating temperature only")
            return self.interp_temp_only(
                temperature, t_idx_min, t_idx_max, -1, wngrid_filter
            )

        # Max temperature
        if check_temperature_max:
            self.debug("Max temperature reached. Interpolating pressure only")
            return self.interp_pressure_only(
                pressure, p_idx_min, p_idx_max, -1, wngrid_filter
            )

        if check_pressure_min:
            self.debug("Min pressure reached. Interpolating temperature only")
            return self.interp_temp_only(
                temperature, t_idx_min, t_idx_max, 0, wngrid_filter
            ).ravel()

        if check_temperature_min:
            self.debug("Min temperature reached. Interpolating pressure only")
            return self.interp_pressure_only(
                pressure, p_idx_min, p_idx_max, 0, wngrid_filter
            ).ravel()

        q_11 = self.xsecGrid[p_idx_min, t_idx_min][wngrid_filter].ravel()
        q_12 = self.xsecGrid[p_idx_min, t_idx_max][wngrid_filter].ravel()
        q_21 = self.xsecGrid[p_idx_max, t_idx_min][wngrid_filter].ravel()
        q_22 = self.xsecGrid[p_idx_max, t_idx_max][wngrid_filter].ravel()

        temperature_max = self.temperatureGrid[t_idx_max]
        temperature_min = self.temperatureGrid[t_idx_min]
        pressure_max = self.logPressure[p_idx_max]
        pressure_min = self.logPressure[p_idx_min]

        if self._interp_mode == "linear":
            return intepr_bilin(
                q_11,
                q_12,
                q_21,
                q_22,
                temperature,
                temperature_min,
                temperature_max,
                pressure,
                pressure_min,
                pressure_max,
            )
        elif self._interp_mode == "exp":
            return interp_exp_and_lin(
                q_11,
                q_12,
                q_21,
                q_22,
                temperature,
                temperature_min,
                temperature_max,
                pressure,
                pressure_min,
                pressure_max,
            )
        else:
            raise ValueError(f"Unknown interpolation mode {self._interp_mode}")

    def compute_opacity(
        self,
        temperature: float,
        pressure: float,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> npt.NDArray[np.float64]:
        """Compute opacity.

        Interpolates the opacity grid to the given temperature and pressure.

        Parameters
        ----------
        temperature: float
            Temperature in K
        pressure: float
            Pressure in Pa
        wngrid: :obj:`array`, optional
            Wavenumber grid to restrict to.

        Returns
        -------
        :obj:`array`
            Interpolated opacity.

        """
        import math

        logpressure = math.log10(pressure)
        return (
            self.interp_bilinear_grid(
                temperature,
                logpressure,
                *self.find_closest_index(temperature, logpressure),
                wngrid,
            )
            / 10000
        )



"""Opacity for hydrogen ion."""
import typing as t
import warnings

import numpy as np
import numpy.typing as npt

from taurex.exceptions import InvalidModelException
from taurex.model import OneDForwardModel
from taurex.output import OutputGroup

from .contribution import Contribution


class HydrogenIon(Contribution):
    """Class to build opacity for hydrogen ion."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__("HydrogenIon")
        self.photo_thresh = 1.6419
        self.alpha = 1.438776877e-2 * 1e6

    def f(self, lamb: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Dont remember what this was."""
        c_n = np.array([1, 52.519, 49.534, -118.858, 92.536, -34.194, 4.982])
        n = np.arange(c_n.shape[0])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            to_ret = np.sum(
                c_n[..., None]
                * ((1 / lamb[None, ...] - 1 / self.photo_thresh) ** (n[..., None] / 2)),
                axis=0,
            )
        return to_ret
        # return np.nan_to_num(sum([c * ((1 / lamb - 1 / self.photo_thresh) **
        # (n / 2)) for n, c in enumerate(C_n)]))

    def sigma_pd(self, lamb: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Dont remember what this was."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            to_ret = np.nan_to_num(
                1e-18
                * (lamb**3)
                * (1 / lamb - 1 / self.photo_thresh) ** (3 / 2)
                * self._f_res
            )
        return to_ret

    def k_bf(
        self, lamb: npt.NDArray[np.float64], temperature: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Compute bound-free."""
        return (
            0.750
            * temperature ** (-5 / 2)
            * np.exp(self.alpha / (self.photo_thresh * temperature))
            * (1 - np.exp(-self.alpha / (lamb * temperature)))
            * self.sigma_pd(lamb)
        )

    def precalc_k_ff(self, lamb: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Precalculate free-free (For performance)."""
        c_3a = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [2483.3460, 285.8270, -2054.2910, 2827.7760, -1341.5370, 208.9520],
                [-3449.8890, -1158.3820, 8746.5230, -11485.6320, 5303.6090, -812.9390],
                [2200.0400, 2427.7190, -13651.1050, 16755.5240, -7510.4940, 1132.7380],
                [-696.2710, -1841.4000, 8642.9700, -10051.5300, 4400.0670, -655.0200],
                [88.2830, 444.5170, -1863.8640, 2095.2880, -901.7880, 132.9850],
            ]
        )

        c_3b = np.array(
            [
                [518.1021, -734.8666, 1021.1775, -479.0721, 93.1373, -6.4285],
                [473.2636, 1443.4137, -1977.3395, 922.3575, -178.9275, 12.3600],
                [-482.2089, -737.1616, 1096.8827, -521.1341, 101.7963, -7.0571],
                [115.5291, 169.6374, -245.6490, 114.2430, -21.9972, 1.5097],
            ]
        )

        self.threea_indices = np.where(lamb > 0.3645)[0]
        self.threeb_indices = np.where((0.1823 < lamb) & (lamb <= 0.3645))[0]

        self.ka, self.pa = self.pre_k_ff_coeff(lamb[self.threea_indices], c_3a)
        self.kb, self.pb = self.pre_k_ff_coeff(lamb[self.threeb_indices], c_3b)

    def k_ff(
        self, lamb: npt.NDArray[np.float64], temperature: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Compute free-free."""
        res = np.zeros_like(lamb)
        res[self.threea_indices] = (
            np.sum(((5040 / temperature) ** self.pa) * self.ka, axis=0) * 1e-29
        )
        res[self.threeb_indices] = (
            np.sum(((5040 / temperature) ** self.pb) * self.kb, axis=0) * 1e-29
        )
        return res

    def pre_k_ff_coeff(
        self, lamb: npt.NDArray[np.float64], coeff: npt.NDArray[np.float64]
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Get coefficients for k for free-free absorption."""
        # k = np.zeros_like(lamb)
        a_coeff = coeff[..., 0]
        b_coeff = coeff[..., 1]
        c_coeff = coeff[..., 2]
        d_coeff = coeff[..., 3]
        e_coeff = coeff[..., 4]
        f_coeff = coeff[..., 5]

        n = np.arange(coeff.shape[0])

        lamb_2 = lamb[None, ...]

        k = (
            (lamb_2**2) * a_coeff[..., None]
            + b_coeff[..., None]
            + c_coeff[..., None] / lamb_2
            + d_coeff[..., None] / (lamb_2**2)
            + e_coeff[..., None] / (lamb_2**3)
            + f_coeff[..., None] / (lamb_2**4)
        )
        ppart = (n[..., None] + 2) / 2
        return k, ppart
        # return 1e-29 * k

    def k_ff_coeff(
        self,
        lamb: npt.NDArray[np.float64],
        temperature: npt.NDArray[np.float64],
        coeff: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """Get coefficients for k for free-free absorption."""
        # k = np.zeros_like(lamb)
        a_coeff = coeff[..., 0]
        b_coeff = coeff[..., 1]
        c_coeff = coeff[..., 2]
        d_coeff = coeff[..., 3]
        e_coeff = coeff[..., 4]
        f_coeff = coeff[..., 5]

        n = np.arange(coeff.shape[0])

        lamb_2 = lamb[None, ...]

        k = np.sum(
            ((5040 / temperature) ** ((n[..., None] + 2) / 2))
            * (
                (lamb_2**2) * a_coeff[..., None]
                + b_coeff[..., None]
                + c_coeff[..., None] / lamb_2
                + d_coeff[..., None] / (lamb_2**2)
                + e_coeff[..., None] / (lamb_2**3)
                + f_coeff[..., None] / (lamb_2**4)
            ),
            axis=0,
        )

        return 1e-29 * k

    def prepare_each(
        self, model: OneDForwardModel, wngrid
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Computes and weighs cross-section for a H-"""
        self._nlayers = model.nLayers
        self._P_dyne = model.pressureProfile * 1e6

        # Since our pressure is in Pa already:
        self._P_dyne = model.pressureProfile * 1e1
        self._temperature_profile = model.temperatureProfile

        if "H" not in model.chemistry.activeGases + model.chemistry.inactiveGases:
            raise InvalidModelException("No hydrogen in chemical model")

        if "e-" not in model.chemistry.activeGases + model.chemistry.inactiveGases:
            raise InvalidModelException("No free electrons in chemical model")

        # Taking the H and e- profiles from chemistry class.
        # This allows flexibility in the profiles ;)
        self._hydrogen_mixratio = model.chemistry.get_gas_mix_profile("H")
        self.debug("hydrogen %s", self._hydrogen_mixratio)
        self._electron_mixratio = model.chemistry.get_gas_mix_profile("e-")
        self.debug("e- %s", self._electron_mixratio)

        xsec_ff = []
        xsec_bf = []
        self.sigma_xsec = np.zeros(shape=(self._nlayers, wngrid.shape[0]))
        self._f_res = self.f(10000 / wngrid)
        self.precalc_k_ff(10000 / wngrid)
        for i in range(self._nlayers):
            xsec_ff = (
                self.k_ff(10000 / wngrid, self._temperature_profile[i])
                * self._P_dyne[i]
                * self._hydrogen_mixratio[i]
                * self._electron_mixratio[i]
            )
            xsec_bf = (
                self.k_bf(10000 / wngrid, self._temperature_profile[i])
                * self._P_dyne[i]
                * self._hydrogen_mixratio[i]
                * self._electron_mixratio[i]
            )

            self.sigma_xsec[i, :] = xsec_ff[:] + xsec_bf[:]

        # convert cm2 to m2 ??
        self.sigma_xsec *= 1e-4

        self.debug("final xsec %s", self.sigma_xsec)
        self.debug("final xsec %s", self.sigma_xsec.max())

        # self._total_contrib[...]=0.0
        yield "HydrogenIon", self.sigma_xsec

    @property
    def sigma(self) -> npt.NDArray[np.float64]:
        """Cross-section."""
        return self.sigma_xsec

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write values to file."""
        contrib = super().write(output)
        contrib.write_scalar("hydrogen", self._hydrogen_mixratio)
        contrib.write_scalar("e-", self._electron_mixratio)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return input keywords for the contribution."""
        return ("HydrogenIon",)

    BIBTEX_ENTRIES = [
        """
        @ARTICLE{John88,
       author = {{John}, T.~L.},
                title = "{Continuous absorption by the
                            negative hydrogen ion reconsidered}",
            journal = {\aap},
            keywords = {Absorption Spectra, Astronomical Spectroscopy, Hydrogen Ions,
            Negative Ions, Photodetachment, Absorptivity,
            Electron Transitions, Wave Functions, Astrophysics},
                year = 1988,
                month = mar,
            volume = {193},
            number = {1-2},
                pages = {189-192},
            adsurl = {https://ui.adsabs.harvard.edu/abs/1988A&A...193..189J},
            adsnote = {Provided by the SAO/NASA Astrophysics Data System}
        }
        """
    ]



"""Basic forward model."""

import typing as t

import numpy as np
import numpy.typing as npt
from astropy import units as u

from taurex.binning import Binner
from taurex.chemistry import Chemistry
from taurex.output import OutputGroup
from taurex.planet import Planet
from taurex.pressure import PressureProfile
from taurex.stellar import Star
from taurex.temperature import TemperatureProfile
from taurex.types import ModelOutputType
from taurex.util import clip_native_to_wngrid

from .model import ForwardModel

if t.TYPE_CHECKING:
    from taurex.contributions import Contribution
else:
    Contribution = object


class SimpleForwardModel(ForwardModel):
    """A 'simple' 1D base model.

    It in the sense that its just
    a fairly standard single profiles model.
    It will handle settingup and initializing, building
    collecting parameters from given profiles for you.
    The only method that requires implementation is:

    - :func:`path_integral`

    """

    WARN = True

    def __init__(
        self,
        name: str,
        planet: t.Optional[Planet] = None,
        star: t.Optional[Star] = None,
        pressure_profile: t.Optional[PressureProfile] = None,
        temperature_profile: t.Optional[TemperatureProfile] = None,
        chemistry: t.Optional[Chemistry] = None,
        nlayers: t.Optional[int] = 100,
        atm_min_pressure: t.Optional[float] = 1e-4,
        atm_max_pressure: t.Optional[float] = 1e6,
        contributions: t.Optional[t.List[Contribution]] = None,
    ):
        """Initialize a 1D forward model.


        Parameters
        ----------
        name: str
            Name to use in logging

        planet:
            Planet model, default planet is Jupiter

        star:
            Star model, default star is Sun-like

        pressure_profile:
            Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
            and ``atm_max_pressure``

        temperature_profile:
            Temperature model, default is an
            :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
            profile at 1500 K

        chemistry:
            Chemistry model, default is
            :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry`
            with ``H2O`` and ``CH4``

        nlayers: int, optional
            Number of layers. Used if ``pressure_profile`` is not defined.

        atm_min_pressure: float, optional
            Pressure at TOA. Used if ``pressure_profile`` is not defined.

        atm_max_pressure: float, optional
            Pressure at BOA. Used if ``pressure_profile`` is not defined.

        """
        import warnings

        super().__init__(name)

        self._planet = planet
        self._star = star
        self._pressure_profile = pressure_profile
        self._temperature_profile = temperature_profile
        self._chemistry = chemistry
        self.debug(
            "Passed: %s %s %s %s %s",
            planet,
            star,
            pressure_profile,
            temperature_profile,
            chemistry,
        )
        self.altitude_profile = None
        self.scaleheight_profile = None
        self.gravity_profile = None
        self._setup_defaults(nlayers, atm_min_pressure, atm_max_pressure)

        self._initialized = False

        self._sigma_opacities = None

        self._native_grid = None

        if contributions:
            for contrib in contributions:
                self.add_contribution(contrib)

        if self.WARN:
            warnings.warn(
                "SimpleForwardModel is deprecated. Use OneDForwardModel instead",
                DeprecationWarning,
                stacklevel=2,
            )

        self.built = False

    def _compute_inital_mu(self):
        """Compute an initial molecular profile."""
        import warnings

        from taurex.data.profiles.chemistry import ConstantGas, TaurexChemistry

        warnings.warn(
            "This method is deprecated and will be removed in a "
            "future version as altitude is never needed.",
            DeprecationWarning,
            stacklevel=2,
        )

        tc = TaurexChemistry()
        tc.addGas(ConstantGas("H2O"))
        self._inital_mu = tc

    def _setup_defaults(
        self, nlayers: int, atm_min_pressure: float, atm_max_pressure: float
    ) -> None:
        """Setup default profiles if not defined.

        Parameters
        ----------
        nlayers: int
            Number of layers

        atm_min_pressure: float
            Pressure at TOA

        atm_max_pressure: float
            Pressure at BOA




        """
        if self._pressure_profile is None:
            from taurex.data.profiles.pressure import SimplePressureProfile

            self.info(
                "No pressure profile defined, using simple pressure " "profile with"
            )
            self.info(
                "parameters nlayers: %s, atm_pressure_range=(%s,%s)",
                nlayers,
                atm_min_pressure,
                atm_max_pressure,
            )

            self._pressure_profile = SimplePressureProfile(
                nlayers, atm_min_pressure, atm_max_pressure
            )

        if self._planet is None:
            from taurex.data import Planet

            self.warning("No planet defined, using Jupiter as planet")
            self._planet = Planet()

        if self._temperature_profile is None:
            from taurex.data.profiles.temperature import Isothermal

            self.warning(
                "No temeprature profile defined using default "
                "Isothermal profile with T=1500 K"
            )
            self._temperature_profile = Isothermal()

        if self._chemistry is None:
            from taurex.data.profiles.chemistry import ConstantGas, TaurexChemistry

            tc = TaurexChemistry()
            self.warning(
                "No gas profile set, using constant profile with H2O " "and CH4"
            )
            tc.addGas(ConstantGas("H2O", mix_ratio=1e-5))
            tc.addGas(ConstantGas("CH4", mix_ratio=1e-6))
            self._chemistry = tc

        if self._star is None:
            from taurex.data.stellar import BlackbodyStar

            self.warning("No star, using the Sun")
            self._star = BlackbodyStar()

    def initialize_profiles(self) -> None:
        """Initializes all profiles."""
        self.info("Computing pressure profile")

        self.pressure.compute_pressure_profile()

        self._temperature_profile.initialize_profile(
            self._planet, self.pressure.nLayers, self.pressure.profile
        )

        # Initialize the atmosphere with a constant gas profile
        # if self._initialized is False:
        #     self._inital_mu.initialize_chemistry(
        #         self.pressure.nLayers,
        #         self.temperatureProfile,
        #         self.pressureProfile,
        #         None,
        #     )

        #     self._compute_altitude_gravity_scaleheight_profile(
        #         self._inital_mu.muProfile
        #     )

        #     self._initialized = True

        # Setup any photochemistry
        self._chemistry.set_star_planet(self.star, self.planet)

        # Now initialize the gas profile real
        self._chemistry.initialize_chemistry(
            self.pressure.nLayers,
            self.temperatureProfile,
            self.pressureProfile,
            None,
            # self.altitude_profile,
        )

        # Compute gravity scale height
        self._compute_altitude_gravity_scaleheight_profile()

    def collect_fitting_parameters(self) -> None:
        """Combine all fitting parameters from components."""
        self._fitting_parameters = {}
        self._fitting_parameters.update(self.fitting_parameters())
        self._fitting_parameters.update(self._planet.fitting_parameters())
        if self._star is not None:
            self._fitting_parameters.update(self._star.fitting_parameters())
        self._fitting_parameters.update(self.pressure.fitting_parameters())

        self._fitting_parameters.update(self._temperature_profile.fitting_parameters())

        self._fitting_parameters.update(self._chemistry.fitting_parameters())

        for contrib in self.contribution_list:
            self._fitting_parameters.update(contrib.fitting_parameters())

        self.debug(
            "Available Fitting params: %s", list(self._fitting_parameters.keys())
        )

    def collect_derived_parameters(self):
        """Combine all derived parameters from all components."""
        self._derived_parameters = {}
        self._derived_parameters.update(self.derived_parameters())
        self._derived_parameters.update(self._planet.derived_parameters())
        if self._star is not None:
            self._derived_parameters.update(self._star.derived_parameters())
        self._derived_parameters.update(self.pressure.derived_parameters())

        self._derived_parameters.update(self._temperature_profile.derived_parameters())

        self._derived_parameters.update(self._chemistry.derived_parameters())

        for contrib in self.contribution_list:
            self._derived_parameters.update(contrib.derived_parameters())

        self.debug(
            "Available derived params: %s", list(self._derived_parameters.keys())
        )

    def build(self) -> None:
        """Build the forward model.

        Must be called at least once before running :func:`model`

        Will automatically be called by :func:`model` if not called before.
        """

        self.contribution_list.sort(key=lambda x: x.order)

        self.info("Building model........")
        # self._compute_inital_mu()
        self.info("Collecting paramters")
        self.collect_fitting_parameters()
        self.collect_derived_parameters()
        self.info("Setting up profiles")
        # self.initialize_profiles()

        self.info("Setting up contributions")
        # for contrib in self.contribution_list:
        #     contrib.build(self)
        self.info("DONE")
        self.built = True

    # altitude, gravity and scale height profile
    def _compute_altitude_gravity_scaleheight_profile(
        self, mu_profile: t.Optional[npt.NDArray[np.float64]] = None
    ) -> None:
        """Computes altitude, gravity and scale height of the atmosphere.
        Only call after :func:`build` has been called at least once.

        Parameters
        ----------
        mu_profile, optional:
            Molecular weight profile at each layer

        """

        # from taurex.constants import KBOLTZ
        if mu_profile is None:
            mu_profile = self._chemistry.muProfile

        pressure_levels = self.pressure.pressure_profile_levels
        z, scaleheight, g, deltaz = self.planet.calculate_scale_properties(
            self.temperatureProfile, pressure_levels, mu_profile
        )
        self.altitude_profile = z[:-1]
        self.scaleheight_profile = scaleheight[:-1]
        self.gravity_profile = g[:-1]
        self.altitude_boundaries = z
        self.deltaz = deltaz

    @property
    def pressureProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Central atmospheric pressure profile in Pa."""
        return self.pressure.profile

    @property
    def temperatureProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Atmospheric temperature profile in K."""
        return self._temperature_profile.profile

    @property
    def densityProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Atmospheric density profile in m-3."""
        from taurex.constants import KBOLTZ

        return (self.pressureProfile) / (KBOLTZ * self.temperatureProfile)

    @property
    def altitudeProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Atmospheric altitude profile in m."""
        return self.altitude_profile

    @property
    def nLayers(self) -> int:  # noqa: N802
        """Number of layers."""
        return self.pressure.nLayers

    @property
    def chemistry(self) -> Chemistry:
        """Chemistry model."""
        return self._chemistry

    @property
    def pressure(self) -> PressureProfile:
        """Pressure model."""
        return self._pressure_profile

    @property
    def temperature(self) -> TemperatureProfile:
        """Temperature model."""
        return self._temperature_profile

    @property
    def star(self) -> Star:
        """Star model."""
        return self._star

    @property
    def planet(self) -> Planet:
        """Planet model."""
        return self._planet

    def set_native_grid(
        self, spectral_grid: t.Union[u.Quantity, npt.NDArray[np.float64]]
    ) -> None:
        """Sets the native grid.

        Parameters
        ----------
        wngrid:
            Wavenumber grid
        """
        if isinstance(spectral_grid, u.Quantity):
            wngrid = spectral_grid.to(u.k, equivalencies=u.spectral()).value

        wngrid = np.array(wngrid)
        # Sort the grid
        wngrid = np.sort(wngrid)

        self._native_grid = wngrid

    def auto_grid(self) -> None:
        """Automatically sets the native grid."""
        self._native_grid = None

    @property
    def nativeWavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Best grid for given cross-sections.

        Searches through active molecules to determine the
        native wavenumber grid

        Returns
        -------

        wngrid: :obj:`array`
            Native grid

        Raises
        ------
        InvalidModelException
            If no active molecules in atmosphere
        """
        from taurex.cache import GlobalCache
        from taurex.cache.opacitycache import OpacityCache
        from taurex.exceptions import InvalidModelException

        if self._native_grid is not None:
            return self._native_grid
        cacher = OpacityCache()

        if GlobalCache()["opacity_method"] == "ktables":
            from taurex.cache.ktablecache import KTableCache

            cacher = KTableCache()

        active_gases = self.chemistry.activeGases

        wavenumbergrid = [cacher[gas].wavenumberGrid for gas in active_gases]

        current_grid = None
        for wn in wavenumbergrid:
            if current_grid is None:
                current_grid = wn
            if wn.shape[0] > current_grid.shape[0]:
                current_grid = wn

        if current_grid is None:
            self.error("No active molecules detected")
            self.error("Most likely no cross-sections/ktables were detected")
            raise InvalidModelException("No active absorbing molecules")

        return current_grid

    def model(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> ModelOutputType:
        """Runs the forward model.

        Parameters
        ----------

        wngrid:
            Wavenumber grid, default is to use native grid

        cutoff_grid:
            Run model only on ``wngrid`` given, default is ``True``

        Returns
        -------

        native_grid:
            Native wavenumber grid, clipped if ``wngrid`` passed

        depth:
            Resulting depth

        tau:
            Optical depth.

        extra: ``None``
            Empty
        """
        if not self.built:
            self.build()
        # Setup profiles
        self.initialize_profiles()

        # Clip grid if necessary
        native_grid = self.nativeWavenumberGrid
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        # Initialize star
        self._star.initialize(native_grid)

        # Prepare contributions
        for contrib in self.contribution_list:
            contrib.prepare(self, native_grid)

        # Compute path integral
        absorp, tau = self.path_integral(native_grid, False)

        return native_grid, absorp, tau, None

    def model_contrib(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        t.Dict[
            str,
            t.Tuple[
                npt.NDArray[np.float64], npt.NDArray[np.float64], t.Optional[t.Any]
            ],
        ],
    ]:
        """Models each contribution seperately."""
        # Setup profiles
        self.initialize_profiles()

        # Copy over contribution list
        full_contrib_list = self.contribution_list
        # Get the native grid
        native_grid = self.nativeWavenumberGrid

        # Clip grid
        all_contrib_dict = {}
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        # Initialize star
        self._star.initialize(native_grid)

        for contrib in full_contrib_list:
            self.contribution_list = [contrib]
            contrib.prepare(self, native_grid)
            absorp, tau = self.path_integral(native_grid, False)
            all_contrib_dict[contrib.name] = (absorp, tau, None)

        self.contribution_list = full_contrib_list
        return native_grid, all_contrib_dict

    def model_full_contrib(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        t.Dict[
            str,
            t.List[
                t.Tuple[
                    str,
                    npt.NDArray[np.float64],
                    npt.NDArray[np.float64],
                    t.Optional[t.Any],
                ]
            ],
        ],
    ]:
        """Model each sub-component of each contribution.

        Like :func:`model_contrib` except all components for
        each contribution are modelled

        e.g. Absorption will be modelled seperately for each molecule
        in the atmosphere.

        """
        native_grid = self.nativeWavenumberGrid
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        self.initialize_profiles()
        self._star.initialize(native_grid)

        result_dict = {}

        full_contrib_list = self.contribution_list

        self.debug("NATIVE GRID %s", native_grid.shape)

        self.info("Modelling each contribution.....")
        for contrib in full_contrib_list:
            self.contribution_list = [contrib]
            contrib_name = contrib.name
            contrib_res_list = []

            for name, __ in contrib.prepare_each(self, native_grid):
                self.info("\t%s---%s contribtuion", contrib_name, name)
                absorp, tau = self.path_integral(native_grid, False)
                contrib_res_list.append((name, absorp, tau, None))

            result_dict[contrib_name] = contrib_res_list

        self.contribution_list = full_contrib_list
        return native_grid, result_dict

    def compute_error(
        self,
        samples: t.Callable[[], float],
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        binner: t.Optional[Binner] = None,
    ) -> t.Tuple[
        t.Dict[str, npt.NDArray[np.float64]], t.Dict[str, npt.NDArray[np.float64]]
    ]:
        """Computes standard deviations from samples.

        Parameters
        ----------

        samples:
            A callable function that returns a weight for each sample

        wngrid:
            Wavenumber grid, default is to use native grid

        binner:
            A :class:`~taurex.binning.binner.Binner` object to bin the spectrum

        """
        from taurex.util.math import OnlineVariance

        tp_profiles = OnlineVariance()
        active_gases = OnlineVariance()
        inactive_gases = OnlineVariance()
        has_condensates = self.chemistry.hasCondensates

        cond = OnlineVariance() if has_condensates else None

        binned_spectrum = OnlineVariance() if binner is not None else None
        native_spectrum = OnlineVariance()

        for weight in samples():
            native_grid, native, tau, _ = self.model(wngrid=wngrid, cutoff_grid=False)

            tp_profiles.update(self.temperatureProfile, weight=weight)
            active_gases.update(self.chemistry.activeGasMixProfile, weight=weight)
            inactive_gases.update(self.chemistry.inactiveGasMixProfile, weight=weight)

            if cond is not None:
                cond.update(self.chemistry.condensateMixProfile, weight=weight)

            native_spectrum.update(
                np.maximum(np.nan_to_num(native), 1e-20), weight=weight
            )

            if binned_spectrum is not None:
                binned = np.maximum(
                    np.nan_to_num(binner.bindown(native_grid, native)[1]), 1e-20
                )

                binned_spectrum.update(binned, weight=weight)

        tp_std = np.sqrt(tp_profiles.parallelVariance())
        active_std = np.sqrt(active_gases.parallelVariance())
        inactive_std = np.sqrt(inactive_gases.parallelVariance())

        profile_dict = {
            "temp_profile_std": tp_std,
            "active_mix_profile_std": active_std,
            "inactive_mix_profile_std": inactive_std,
        }

        if cond is not None:
            profile_dict["condensate_profile_std"] = np.sqrt(cond.parallelVariance())

        spectrum_dict = {"native_std": np.sqrt(native_spectrum.parallelVariance())}
        if binned_spectrum is not None:
            spectrum_dict["binned_std"] = np.sqrt(binned_spectrum.parallelVariance())

        return profile_dict, spectrum_dict

    def path_integral(
        self, wngrid: npt.NDArray[np.float64], return_contrib: t.Optional[bool]
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Main integration function.

        Must return the absorption and optical depth for the given wngrid.

        """
        raise NotImplementedError

    def generate_profiles(self) -> None:
        """Generate profiles to store."""
        from taurex.util.output import generate_profile_dict

        prof = generate_profile_dict(self)
        prof["mu_profile"] = self.chemistry.muProfile
        return prof

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write forward model to output group.

        Will also write all components to the output group.

        """
        # Run a model if needed
        self.model()

        model = super().write(output)

        self._chemistry.write(model)
        self._temperature_profile.write(model)
        self.pressure.write(model)
        self._planet.write(model)
        self._star.write(model)
        return model

    def citations(self) -> str:
        """Citations for model."""
        from taurex.cache import OpacityCache
        from taurex.data.citation import unique_citations_only

        model_citations = super().citations()
        model_citations.extend(self.chemistry.citations())
        model_citations.extend(self.temperature.citations())
        model_citations.extend(self.pressure.citations())
        model_citations.extend(self.planet.citations())
        model_citations.extend(self.star.citations())

        # Get cache citations
        active_gases = self.chemistry.activeGases

        opacitycache = OpacityCache()

        # Do cross sections

        for g in active_gases:
            try:
                xsec = opacitycache.opacity_dict[g]
                model_citations.extend(xsec.citations())
            except KeyError:
                continue
        return unique_citations_only(model_citations)


# A much better alias for the class
class OneDForwardModel(SimpleForwardModel):
    """A forward model for a 1D atmosphere.

    Automatically sets up the following profiles:
        - Temperature
        - Pressure
        - Chemistry
        - Planet
        - Star
        - Contributions


    Must implement the following methods:
        - :func:`path_integral`
            - Main integration function that must
            return the absorption and optical depth
            for the given wngrid.

    """

    pass



"""Direct imaging model."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.constants import PI

from .emission import EmissionModel


class DirectImageModel(EmissionModel):
    """A forward model for direct imaging of exo-planets."""

    def compute_final_flux(
        self, f_total: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Compute the final flux.

        This is the emission flux that is observed at the telescope directly
        from an exo-planet.

        """
        star_distance_meters = self._star.distance * 3.08567758e16

        sdr = pow((star_distance_meters / 3.08567758e16), 2)
        sdr = 1.0
        planet_radius = self._planet.fullRadius

        return (
            (f_total * (planet_radius**2) * 2.0 * PI)
            / (4 * PI * (star_distance_meters**2))
        ) * sdr

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return (
            "direct",
            "directimage",
        )



"""Math functions for Numba."""
# flake8: noqa
import math

import numba
import numpy as np
from numba import float64, vectorize


@numba.vectorize([float64(float64, float64, float64, float64)])
def _expstage0(x1, x2, x11, x21):
    return x1 * x11 - x2 * (x11 - x21)


@numba.vectorize([float64(float64, float64)], fastmath=True)
def _expstage1(x1, x2):
    return math.log(x1 / x2)


@numba.vectorize([float64(float64, float64)], fastmath=True)
def _expstage2(C, x):
    return C * x


@numba.vectorize([float64(float64, float64, float64)], fastmath=True)
def _expstage3(C, x1, x2):
    return C * x1 * x2


@numba.njit(nogil=True, fastmath=True)
def interp_exp_and_lin_broken(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    res = np.zeros_like(x11)
    x0 = -Pmin
    x1 = Pmax + x0
    x2 = P + x0
    factor1 = 1.0 / (T * (Tmax - Tmin))
    factor2 = 1.0 / x1
    x3 = _expstage0(x1, x2, x11, x21)
    x4 = _expstage0(x1, x2, x12, x22)
    x5 = _expstage1(x3, x4)
    x6 = _expstage2(Tmax * (-T + Tmin) * factor1, x5)
    x7 = _expstage3(factor2, x3, x6)
    for i in range(x11.shape[0]):
        res[i] = x7[i] * math.exp(x6[i])
    return res


@numba.njit
def interp_lin_numba(x11, x12, P, Pmin, Pmax):
    # return (x11*(Pmax - Pmin) - (P - Pmin)*(x11 - x12))/(Pmax - Pmin)
    N = x11.shape[0]
    diff = Pmax - Pmin
    scale = (P - Pmin) / diff
    out = np.zeros_like(x11)
    for n in range(N):
        out[n] = x11[n] - scale * (x11[n] - x12[n])

    return out


@numba.njit
def intepr_bilin_numba_II(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    N = x11.shape[0]
    Pdiff = Pmax - Pmin
    Tdiff = Tmax - Tmin
    Pscale = (P - Pmin) / Pdiff
    Tscale = (T - Tmin) / Tdiff

    out = np.zeros_like(x11)

    for n in range(N):
        out[n] = (
            x11[n]
            - Pscale * (x11[n] - x21[n])
            - Pscale * Tscale * (x21[n] - x11[n] + x12[n] - x22[n])
            - Tscale * (x11[n] - x12[n])
        )

    return out


@numba.vectorize([float64(float64, float64, float64)], fastmath=True)
def _linstage0(x11, x21, x):
    return x * (x11 - x21)


@numba.njit(nogil=True, fastmath=True)
def intepr_bilin_numba(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    x0 = -Pmin
    x1 = Pmax + x0
    x2 = -Tmin
    x3 = Tmax + x2
    x4 = P + x0

    factor = 1.0 / (x1 * x3)

    x5 = _linstage0(x11, x21, x4)
    x6 = _linstage0(x11, x12, x1)
    x7 = _linstage0(x12, x22, x4)
    res = np.zeros_like(x11)
    for i in range(x11.shape[0]):
        res[i] = (
            x1 * x11[i] * x3 - x3 * x5[i] - (T + x2) * (x6[i] + x7[i] - x5[i])
        ) * factor

    return res


@numba.njit(fastmath=True)
def interp_exp_numba(x11, x12, temperature, temperature_min, temperature_max):
    N = x11.shape[0]
    scale = (
        temperature_max
        * (temperature_min - temperature)
        / (temperature_max - temperature_min)
    )
    out = np.zeros_like(x11)
    for n in range(N):
        out[n] = x11[n] * np.exp(scale * np.log(x11[n] / x12[n]))

    return out



"""Optimized Math functions used in taurex"""


import typing as t

import numpy as np
import numpy.typing as npt

from taurex.log import setup_log
from taurex.types import AnyValType

_log = setup_log(__name__)

try:
    from .math_numba import intepr_bilin_numba_II, interp_lin_numba

    numba_enabled = True
except ImportError:
    _log.warning("Numba not installed, using numpy")
    numba_enabled = False


def interp_exp_and_lin_numpy(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    x21: npt.NDArray[np.float64],
    x22: npt.NDArray[np.float64],
    temperature: float,
    temperature_min: float,
    temperature_max: float,
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    """2D interpolation.

    Applies linear interpolation across pressure and e interpolation
    across temperature between pressure_min,pressure_max and
    temperature_min,temperature_max

    Parameters
    ----------
    x11: array
        Array corresponding to pressure_min,temperature_min

    x12: array
        Array corresponding to pressure_min,temperature_max

    x21: array
        Array corresponding to pressure_max,temperature_min

    x22: array
        Array corresponding to pressure_max,temperature_max

    temperature: float
        Coordinate to exp interpolate to

    temperature_min: float
        Nearest known temperature coordinate where temperature_min < temperature

    temperature_max: float
        Nearest known temperature coordinate where temperature < temperature_max

    pressure: float
        Coordinate to linear interpolate to

    pressure_min: float
        Nearest known pressure coordinate where pressure_min < pressure

    pressure_max: float
        Nearest known pressure coordinate where pressure < pressure_max

    """

    return (
        (x11 * (pressure_max - pressure_min) - (pressure - pressure_min) * (x11 - x21))
        * np.exp(
            temperature_max
            * (-temperature + temperature_min)
            * np.log(
                (
                    x11 * (pressure_max - pressure_min)
                    - (pressure - pressure_min) * (x11 - x21)
                )
                / (
                    x12 * (pressure_max - pressure_min)
                    - (pressure - pressure_min) * (x12 - x22)
                )
            )
            / (temperature * (temperature_max - temperature_min))
        )
        / (pressure_max - pressure_min)
    )


def interp_exp_numpy(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    temperature,
    temperature_min,
    temperature_max,
) -> npt.NDArray[np.float64]:
    return x11 * np.exp(
        temperature_max
        * (-temperature + temperature_min)
        * np.log(x11 / x12)
        / (temperature * (temperature_max - temperature_min))
    )


def interp_lin_numpy(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    return (
        x11 * (pressure_max - pressure_min) - (pressure - pressure_min) * (x11 - x12)
    ) / (pressure_max - pressure_min)


def interp_bilin_numpy(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    x21: npt.NDArray[np.float64],
    x22: npt.NDArray[np.float64],
    temperature: float,
    temperature_min: float,
    temperature_max: float,
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    pressure_diff = pressure_max - pressure_min
    temperature_diff = temperature_max - temperature_min
    pressure_scale = (pressure - pressure_min) / pressure_diff
    temperature_scale = (temperature - temperature_min) / temperature_diff

    return (
        x11
        - pressure_scale * (x11 - x21)
        - pressure_scale * temperature_scale * (x21 - x11 + x12 - x22)
        - temperature_scale * (x11 - x12)
    )


def intepr_bilin_numexpr(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    x21: npt.NDArray[np.float64],
    x22: npt.NDArray[np.float64],
    temperature: float,
    temperature_min: float,
    temperature_max: float,
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    import numexpr as ne

    return ne.evaluate(
        "(x11*(pressure_max - pressure_min)*(temperature_max - temperature_min)"
        " - (pressure - pressure_min)*(temperature_max - temperature_min)*(x11 - x21)"
        " - (temperature - temperature_min)*(-(pressure - pressure_min)*(x11 - x21)"
        " + (pressure - pressure_min)*(x12 - x22) + (pressure_max - "
        "pressure_min)*(x11 - x12)))/((pressure_max - "
        "pressure_min)*(temperature_max - temperature_min))"
    )


def intepr_bilin_double(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    x21: npt.NDArray[np.float64],
    x22: npt.NDArray[np.float64],
    temperature: float,
    temperature_min: float,
    temperature_max: float,
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    return interp_lin_only(
        interp_lin_only(x11, x12, temperature, temperature_min, temperature_max),
        interp_lin_only(x21, x22, temperature, temperature_min, temperature_max),
        pressure,
        pressure_min,
        pressure_max,
    )


def intepr_bilin_old(
    x11: npt.NDArray[np.float64],
    x12: npt.NDArray[np.float64],
    x21: npt.NDArray[np.float64],
    x22: npt.NDArray[np.float64],
    temperature: float,
    temperature_min: float,
    temperature_max: float,
    pressure: float,
    pressure_min: float,
    pressure_max: float,
) -> npt.NDArray[np.float64]:
    return (
        x11 * (pressure_max - pressure_min) * (temperature_max - temperature_min)
        - (pressure - pressure_min) * (temperature_max - temperature_min) * (x11 - x21)
        - (temperature - temperature_min)
        * (
            -(pressure - pressure_min) * (x11 - x21)
            + (pressure - pressure_min) * (x12 - x22)
            + (pressure_max - pressure_min) * (x11 - x12)
        )
    ) / ((pressure_max - pressure_min) * (temperature_max - temperature_min))


def compute_rayleigh_cross_section(
    wngrid: npt.NDArray[np.float64],
    n: float,
    n_air: t.Optional[float] = 2.6867805e25,
    king: t.Optional[float] = 1.0,
) -> npt.NDArray[np.float64]:
    """Compute Rayleigh cross section."""
    wlgrid = (10000 / wngrid) * 1e-6

    n_factor = (n**2 - 1) / (n_air * (n**2 + 2))
    sigma = 24.0 * (np.pi**3) * king * (n_factor**2) / (wlgrid**4)

    return sigma


def test_nan(val: t.Union[float, npt.ArrayLike]) -> bool:
    """Test if a value is nan."""
    if hasattr(val, "__len__"):
        try:
            return np.isnan(val).any()
        except TypeError:
            # print(type(val))
            return True
    else:
        return val != val


# Choose the best functions for the task
if numba_enabled:
    interp_lin_only = interp_lin_numba
    intepr_bilin = intepr_bilin_numba_II
else:
    interp_lin_only = interp_lin_numpy
    intepr_bilin = interp_bilin_numpy
interp_exp_and_lin = interp_exp_and_lin_numpy
interp_exp_only = interp_exp_numpy


class OnlineVariance:
    """USes the M2 algorithm to compute the variance in a streaming fashion"""

    def __init__(self) -> None:
        """Initialise the class."""
        self.reset()

    def reset(self) -> None:
        """Reset the class."""
        self.count = 0.0
        self.wcount = 0.0
        self.wcount2 = 0.0
        self.mean = None
        self.M2 = None

    def update(self, value: AnyValType, weight: t.Optional[float] = 1.0):
        """Update the variance."""
        self.count += 1
        self.wcount += weight
        self.wcount2 += weight * weight

        if self.mean is None:
            self.mean = value * 0.0
            self.M2 = value * 0.0

        mean_old = self.mean
        try:
            self.mean = mean_old + (weight / self.wcount) * (value - mean_old)
        except ZeroDivisionError:
            self.mean = value * 0.0
        self.M2 += weight * (value - mean_old) * (value - self.mean)

    @property
    def variance(self) -> float:
        """Return the variance."""
        if self.count < 2:
            return np.nan
        else:
            return self.M2 / self.wcount

    @property
    def sampleVariance(self) -> AnyValType:  # noqa: N802
        """Return the sample variance."""
        if self.count < 2:
            return np.nan
        else:
            return self.M2 / (self.wcount - 1)

    # def combine_variance(self,averages, variances, counts):
    #     good_idx = [idx for idx,a in enumerate(averages) if not test_nan(a)]
    #     averages = [averages[idx] for idx in good_idx]
    #     variances = [variances[idx] for idx in good_idx]
    #     counts = [counts[idx] for idx in good_idx]
    #     good_variance = None
    #     if not test_nan(variances):
    #         try:
    #             good_variance = variances[np.where(~np.isnan(variances))[0][0]]*0.0
    #         except IndexError:
    #             good_variance = None
    #     #print(good_idx,'Good',good_variance)
    #     variances = [v if not test_nan(v) else good_variance for v in variances]
    #     #print('NEWAVERAGES',averages)
    #     #print('NEW WEIGHTS',counts)

    #     average = np.average(averages, weights=counts,axis=0)

    #     #print('final average',average)
    #     size = np.sum(counts)

    #     counts = np.array(counts) * size/np.sum(counts)
    #     if hasattr(average,'__len__'):
    #         average = average[None,...]
    #         for x in range(1,len(average.shape)):
    #             counts = counts[:,None]
    #     squares = 0.0
    #     if good_variance is not None:
    #         squares = counts*np.nan_to_num(variances)
    #     #print(counts,variances,squares)
    #     squares = squares + counts*(average - averages)**2

    #     return average,np.sum(squares,axis=0)/size

    def combine_variance(
        self, averages: npt.ArrayLike, variance: npt.ArrayLike, counts: npt.ArrayLike
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Combine different variance calculations together."""
        average = None
        size = np.sum(counts)
        for avg, cnt in zip(averages, counts):
            if cnt == 0:
                continue

            # print('avg',avg)
            if avg is not None and avg is not np.nan:
                if average is None:
                    average = avg * cnt
                else:
                    average += avg * cnt
        average /= size
        # print('AVERGAE',average)
        counts = np.array(counts) * size / np.sum(counts)

        squares = None

        for avg, cnt, var in zip(averages, counts, variance):
            # print('COUNT ',cnt)
            if cnt == 0.0:
                continue
            if cnt > 0.0:
                if squares is None:
                    squares = cnt * (average - avg) ** 2
                else:
                    squares += cnt * (average - avg) ** 2
            if var is not np.nan:
                squares += cnt * var
        # squares = counts*variances
        # squares += counts*(average - averages)**2

        return average, squares / size

    def parallelVariance(self) -> AnyValType:  # noqa: N802
        """Compute the variance in parallel."""
        from taurex import mpi

        variance = self.variance

        mean = self.mean
        if mean is None:
            mean = np.nan

        variances = mpi.allgather(variance)

        averages = mpi.allgather(mean)
        counts = mpi.allgather(self.wcount)
        all_counts = mpi.allgather(self.count)

        if sum(all_counts) < 2:
            return np.nan
        else:
            finalvariance = self.combine_variance(averages, variances, counts)
            return finalvariance[-1]



"""Handles loading of classes from plugins and built-in classes"""

# flake8: noqa:N802

import importlib.metadata
import inspect
import pathlib
import sys
import typing as t
from types import ModuleType

from taurex.core import Singleton
from taurex.log import Logger
from taurex.mixin import Mixin
from taurex.types import PathLike

try:
    EPType = importlib.metadata.EntryPoints
except AttributeError:
    EPType = t.Any


def entry_points(*, group: str) -> EPType:  # type: ignore[name-defined]
    """Wrapper for entry_points to support Python 3.8+ instead of 3.10+."""
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)
    epg = importlib.metadata.entry_points()

    return epg.get(group, [])


class InputKeywordMethod(t.Protocol):
    """Protocol for classes that have input_keywords()"""

    @classmethod
    def input_keywords(cls) -> t.Sequence[str]:
        """Returns a list of input keywords"""
        ...


class DiscoverMethod(t.Protocol):
    """Protocol for classes that have discover() methods"""

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Any]]:
        """Returns a list of input keywords"""
        ...


T = t.TypeVar("T", bound=InputKeywordMethod)
AllT = t.TypeVar("AllT")
OpacT = t.TypeVar("OpacT", bound=DiscoverMethod)
MixinT = t.TypeVar("MixinT", bound=Mixin)


class ClassFactory(Singleton):
    """A singleton factory the discovers new classes from plugins."""

    def init(self) -> None:
        self.log = Logger("ClassFactory")

        self.extension_paths = []
        self.reload_plugins()

    def set_extension_paths(
        self,
        paths: t.Optional[t.List[PathLike]] = None,
        reload: t.Optional[bool] = True,
    ) -> None:
        self.extension_paths = [pathlib.Path(p) for p in paths] if paths else []
        if reload:
            self.reload_plugins()

    def reload_plugins(self) -> None:
        """Reload all plugins and built-in classes."""
        self.log.info("Reloading all modules and plugins")
        self.setup_batteries_included()
        self.setup_batteries_included_mixin()
        self.load_plugins()
        self.load_extension_paths()

    def setup_batteries_included_mixin(self):
        """Collect all the mixin classes that are built into TauREx 3."""
        import taurex.mixin as mixins

        self._temp_mixin_klasses = set(self._collect_temperatures_mixin(mixins))
        self._chem_mixin_klasses = set(self._collect_chemistry_mixin(mixins))
        self._gas_mixin_klasses = set(self._collect_gas_mixin(mixins))
        self._press_mixin_klasses = set(self._collect_pressure_mixin(mixins))
        self._planet_mixin_klasses = set(self._collect_planets_mixin(mixins))
        self._star_mixin_klasses = set(self._collect_star_mixin(mixins))
        self._inst_mixin_klasses = set(self._collect_instrument_mixin(mixins))
        self._model_mixin_klasses = set(self._collect_model_mixin(mixins))
        self._obs_mixin_klasses = set(self._collect_observation_mixin(mixins))
        self._contrib_mixin_klasses = set(self._collect_contributions_mixin(mixins))
        self._opt_mixin_klasses = set(self._collect_optimizer_mixin(mixins))
        self._any_mixin_klasses = set(self._collect_classes(mixins, mixins.AnyMixin))

    def setup_batteries_included(self) -> None:
        """Collect all the classes that are built into TauREx 3."""
        from taurex import (
            chemistry,
            contributions,
            instruments,
            model,
            opacity,
            optimizer,
            planet,
            pressure,
            spectrum,
            stellar,
            temperature,
        )
        from taurex.core import priors
        from taurex.opacity import ktables

        self._temp_klasses = set(self._collect_temperatures(temperature))
        self._chem_klasses = set(self._collect_chemistry(chemistry))
        self._gas_klasses = set(self._collect_gas(chemistry))
        self._press_klasses = set(self._collect_pressure(pressure))
        self._planet_klasses = set(self._collect_planets(planet))
        self._star_klasses = set(self._collect_star(stellar))
        self._inst_klasses = set(self._collect_instrument(instruments))
        self._model_klasses = set(self._collect_model(model))
        self._obs_klasses = set(self._collect_observation(spectrum))
        self._contrib_klasses = set(self._collect_contributions(contributions))

        self._opt_klasses = set(self._collect_optimizer(optimizer))
        self._opac_klasses = set(self._collect_opacity(opacity))
        self._ktab_klasses = set(self._collect_ktables(ktables))
        self._prior_klasses = set(self._collect_priors(priors))

    def load_plugin(self, plugin_module: ModuleType) -> None:
        """Load a plugin module and collect all the classes from it."""
        self._temp_klasses.update(self._collect_temperatures(plugin_module))
        self._chem_klasses.update(self._collect_chemistry(plugin_module))
        self._gas_klasses.update(self._collect_gas(plugin_module))
        self._press_klasses.update(self._collect_pressure(plugin_module))
        self._planet_klasses.update(self._collect_planets(plugin_module))
        self._star_klasses.update(self._collect_star(plugin_module))
        self._inst_klasses.update(self._collect_instrument(plugin_module))
        self._model_klasses.update(self._collect_model(plugin_module))
        self._obs_klasses.update(self._collect_observation(plugin_module))
        self._contrib_klasses.update(self._collect_contributions(plugin_module))
        self._opt_klasses.update(self._collect_optimizer(plugin_module))
        self._opac_klasses.update(self._collect_opacity(plugin_module))
        self._prior_klasses.update(self._collect_priors(plugin_module))
        self._ktab_klasses.update(self._collect_ktables(plugin_module))

        self._temp_mixin_klasses.update(self._collect_temperatures_mixin(plugin_module))
        self._chem_mixin_klasses.update(self._collect_chemistry_mixin(plugin_module))
        self._gas_mixin_klasses.update(self._collect_gas_mixin(plugin_module))
        self._press_mixin_klasses.update(self._collect_pressure_mixin(plugin_module))
        self._planet_mixin_klasses.update(self._collect_planets_mixin(plugin_module))
        self._star_mixin_klasses.update(self._collect_star_mixin(plugin_module))
        self._inst_mixin_klasses.update(self._collect_instrument_mixin(plugin_module))
        self._model_mixin_klasses.update(self._collect_model_mixin(plugin_module))
        self._obs_mixin_klasses.update(self._collect_observation_mixin(plugin_module))
        self._contrib_mixin_klasses.update(
            self._collect_contributions_mixin(plugin_module)
        )

        self._opt_mixin_klasses.update(self._collect_optimizer_mixin(plugin_module))

    def discover_plugins(self) -> t.Tuple[t.Dict[str, ModuleType], t.Dict[str, str]]:
        """Discover all the plugins that are available from entry points."""

        plugins = {}
        failed_plugins = {}

        plugins_entrypoint = entry_points(group="taurex.plugins")

        for entry_point in plugins_entrypoint:
            entry_point_name = entry_point.name

            try:
                module: ModuleType = entry_point.load()
            except Exception as e:
                # For whatever reason do not attempt to load the plugin
                self.log.warning("Could not load plugin %s", entry_point_name)
                self.log.warning("Reason: %s", str(e))
                failed_plugins[entry_point_name] = str(e)
                continue

            plugins[entry_point_name] = module

        return plugins, failed_plugins

    def load_plugins(self):
        """Load all the plugins that are available."""
        plugins, failed_plugins = self.discover_plugins()
        self.log.info("----------Plugin loading---------")
        self.log.debug("Discovered plugins %s", plugins.values())

        for k, v in plugins.items():
            self.log.info("Loading %s", k)
            self.load_plugin(v)

    def load_extension_paths(self):
        """Load all the plugins from the extension paths."""
        import importlib.util

        paths = self.extension_paths
        if paths:
            # Make sure they're unique
            all_files = set(
                sum(
                    [p.glob("*.py") for p in paths],
                    [],
                )
            )

            for f in all_files:
                self.info("Loading extensions from %s", f)
                module_name = f.stem
                spec = importlib.util.spec_from_file_location(module_name, f)
                foo = importlib.util.module_from_spec(spec)
                try:
                    spec.loader.exec_module(foo)
                    self.load_plugin(foo)
                except Exception as e:
                    self.log.warning("Could not load extension from file %s", f)
                    self.log.warning("Reason: %s", str(e))

    def _collect_classes(
        self,
        module: ModuleType,
        base_klass: t.Type[t.Union[T, OpacT, MixinT]],
        check_validity: t.Optional[bool] = True,
    ) -> t.List[t.Type[t.Union[T, OpacT, MixinT]]]:
        """Collects all classes that are a subclass of base."""
        klasses = []
        clsmembers = inspect.getmembers(module, inspect.isclass)
        for _, c in clsmembers:
            if issubclass(c, base_klass) and (c is not base_klass):
                self.log.debug(f" Found class {c.__name__}")
            else:
                continue

            if check_validity:
                try:
                    if hasattr(c, "discover"):
                        c.discover()
                    if hasattr(c, "input_keywords"):
                        c.input_keywords()
                    klasses.append(c)
                except (NotImplementedError, AttributeError):
                    self.log.debug(
                        f"Class {c.__name__} does not implement input_keywords"
                    )
                    continue
            else:
                klasses.append(c)

        return klasses

    def _collect_temperatures(self, module: ModuleType):
        from taurex.temperature import TemperatureProfile

        return self._collect_classes(module, TemperatureProfile)

    def _collect_chemistry(self, module: ModuleType):
        from taurex.chemistry import Chemistry

        return self._collect_classes(module, Chemistry)

    def _collect_gas(self, module: ModuleType):
        from taurex.chemistry import Gas

        return self._collect_classes(module, Gas)

    def _collect_pressure(self, module: ModuleType):
        from taurex.pressure import PressureProfile

        return self._collect_classes(module, PressureProfile)

    def _collect_planets(self, module: ModuleType):
        from taurex.planet import BasePlanet

        return self._collect_classes(module, BasePlanet)

    def _collect_star(self, module: ModuleType):
        from taurex.stellar import Star

        return self._collect_classes(module, Star)

    def _collect_instrument(self, module: ModuleType):
        from taurex.instruments import Instrument

        return self._collect_classes(module, Instrument)

    def _collect_model(self, module: ModuleType):
        from taurex.model import ForwardModel, OneDForwardModel, SimpleForwardModel

        return self._collect_classes(module, ForwardModel)

    def _collect_contributions(self, module):
        from taurex.contributions import Contribution

        return self._collect_classes(module, Contribution)

    def _collect_optimizer(self, module):
        from taurex.optimizer import Optimizer

        return self._collect_classes(module, Optimizer)

    def _collect_opacity(self, module):
        from taurex.opacity import Opacity

        return [c for c in self._collect_classes(module, Opacity)]

    def _collect_ktables(self, module):
        from taurex.opacity.ktables import KTable

        return [c for c in self._collect_classes(module, KTable)]

    def _collect_observation(self, module):
        from taurex.spectrum import BaseSpectrum

        return [c for c in self._collect_classes(module, BaseSpectrum)]

    def _collect_priors(self, module):
        from taurex.core.priors import Prior

        # Does not need to be valid since it does not have input_keywords or discover
        return self._collect_classes(module, Prior, check_validity=False)

    # Mixins
    def _collect_temperatures_mixin(self, module):
        from taurex.mixin import TemperatureMixin

        return self._collect_classes(module, TemperatureMixin)

    def _collect_chemistry_mixin(self, module):
        from taurex.mixin import ChemistryMixin

        return self._collect_classes(module, ChemistryMixin)

    def _collect_gas_mixin(self, module):
        from taurex.mixin import GasMixin

        return self._collect_classes(module, GasMixin)

    def _collect_pressure_mixin(self, module):
        from taurex.mixin import PressureMixin

        return self._collect_classes(module, PressureMixin)

    def _collect_planets_mixin(self, module):
        from taurex.mixin import PlanetMixin

        return self._collect_classes(module, PlanetMixin)

    def _collect_star_mixin(self, module):
        from taurex.mixin import StarMixin

        return self._collect_classes(module, StarMixin)

    def _collect_instrument_mixin(self, module):
        from taurex.mixin import InstrumentMixin

        return self._collect_classes(module, InstrumentMixin)

    def _collect_model_mixin(self, module):
        from taurex.mixin import ForwardModelMixin

        return self._collect_classes(module, ForwardModelMixin)

    def _collect_contributions_mixin(self, module):
        from taurex.mixin import ContributionMixin

        return self._collect_classes(module, ContributionMixin)

    def _collect_optimizer_mixin(self, module):
        from taurex.mixin import OptimizerMixin

        return self._collect_classes(module, OptimizerMixin)

    def _collect_observation_mixin(self, module):
        from taurex.mixin import ObservationMixin

        return self._collect_classes(module, ObservationMixin)

    @property
    def class_dict(self) -> t.Dict[t.Type[t.Any], t.Set[t.Type[t.Any]]]:
        """Returns a dictionary of all classes that are available in TauREx 3."""
        from taurex.chemistry import Chemistry, Gas
        from taurex.contributions import Contribution
        from taurex.core.priors import Prior
        from taurex.instruments import Instrument
        from taurex.model import ForwardModel
        from taurex.opacity import Opacity
        from taurex.opacity.ktables import KTable
        from taurex.optimizer import Optimizer
        from taurex.planet import BasePlanet
        from taurex.pressure import PressureProfile
        from taurex.spectrum import BaseSpectrum
        from taurex.stellar import Star
        from taurex.temperature import TemperatureProfile

        klass_dict = {
            TemperatureProfile: self.temperatureKlasses,
            Chemistry: self.chemistryKlasses,
            Gas: self.gasKlasses,
            PressureProfile: self.pressureKlasses,
            BasePlanet: self.planetKlasses,
            Star: self.starKlasses,
            Instrument: self.instrumentKlasses,
            ForwardModel: self.modelKlasses,
            Contribution: self.contributionKlasses,
            Optimizer: self.optimizerKlasses,
            Opacity: self.opacityKlasses,
            KTable: self.ktableKlasses,
            BaseSpectrum: self.observationKlasses,
            Prior: self.priorKlasses,
        }

    def list_from_base(self, klass_type: t.Type[t.Any]) -> t.List[t.Type[t.Any]]:
        """Returns a list of classes that are a subclass of klass_type."""

        return self.class_dict[klass_type]

    @property
    def temperatureKlasses(self):
        """Returns a list of all temperature classes."""
        return self._temp_klasses

    @property
    def chemistryKlasses(self):
        """Returns a list of all chemistry classes."""
        return self._chem_klasses

    @property
    def gasKlasses(self):
        """Returns a list of all gas classes."""
        return self._gas_klasses

    @property
    def pressureKlasses(self):
        """Returns a list of all pressure classes."""
        return self._press_klasses

    @property
    def planetKlasses(self):
        """Returns a list of all planet classes."""
        return self._planet_klasses

    @property
    def starKlasses(self):
        """Returns a list of all star classes."""
        return self._star_klasses

    @property
    def instrumentKlasses(self):
        """Returns a list of all instrument classes."""
        return self._inst_klasses

    @property
    def modelKlasses(self):
        """Returns a list of all model classes."""
        return self._model_klasses

    @property
    def contributionKlasses(self):
        """Returns a list of all contribution classes."""
        return self._contrib_klasses

    @property
    def optimizerKlasses(self):
        """Returns a list of all optimizer classes."""
        return self._opt_klasses

    @property
    def opacityKlasses(self):
        """Returns a list of all opacity classes."""
        return self._opac_klasses

    @property
    def ktableKlasses(self):
        """Returns a list of all ktable classes."""
        return self._ktab_klasses

    @property
    def observationKlasses(self):
        """Returns a list of all observation classes."""
        return self._obs_klasses

    @property
    def priorKlasses(self):
        """Returns a list of all prior classes."""
        return self._prior_klasses

    # Mixins

    @property
    def temperatureMixinKlasses(self):
        """Returns a list of all temperature mixin classes."""
        return self._temp_mixin_klasses

    @property
    def chemistryMixinKlasses(self):
        """Returns a list of all chemistry mixin classes."""
        return self._chem_mixin_klasses

    @property
    def gasMixinKlasses(self):
        """Returns a list of all gas mixin classes."""
        return self._gas_mixin_klasses

    @property
    def pressureMixinKlasses(self):
        """Returns a list of all pressure mixin classes."""
        return self._press_mixin_klasses

    @property
    def planetMixinKlasses(self):
        """Returns a list of all planet mixin classes."""
        return self._planet_mixin_klasses

    @property
    def starMixinKlasses(self):
        """Returns a list of all star mixin classes."""
        return self._star_mixin_klasses

    @property
    def instrumentMixinKlasses(self):
        """Returns a list of all instrument mixin classes."""
        return self._inst_mixin_klasses

    @property
    def modelMixinKlasses(self):
        """Returns a list of all model mixin classes."""
        return self._model_mixin_klasses

    @property
    def contributionMixinKlasses(self):
        """Returns a list of all contribution mixin classes."""
        return self._contrib_mixin_klasses

    @property
    def optimizerMixinKlasses(self):
        """Returns a list of all optimizer mixin classes."""
        return self._opt_mixin_klasses

    @property
    def observationMixinKlasses(self):
        """Returns a list of all observation mixin classes."""
        return self._obs_mixin_klasses

    @property
    def all_klasses(self) -> t.Set[t.Union[T, OpacT, MixinT]]:
        """Returns a set of all classes that are available in TauREx 3."""
        return (
            self.temperatureKlasses
            | self.chemistryKlasses
            | self.gasKlasses
            | self.pressureKlasses
            | self.planetKlasses
            | self.starKlasses
            | self.instrumentKlasses
            | self.modelKlasses
            | self.contributionKlasses
            | self.optimizerKlasses
            | self.opacityKlasses
            | self.ktableKlasses
            | self.observationKlasses
            | self.priorKlasses
            | self.temperatureMixinKlasses
            | self.chemistryMixinKlasses
            | self.gasMixinKlasses
            | self.pressureMixinKlasses
            | self.planetMixinKlasses
            | self.starMixinKlasses
            | self.instrumentMixinKlasses
            | self.modelMixinKlasses
            | self.contributionMixinKlasses
            | self.optimizerMixinKlasses
            | self.observationMixinKlasses
        )

    @property
    def mixin_klasses(self) -> t.Set[t.Type[Mixin]]:
        """Returns a set of all mixin classes that are available in TauREx 3."""
        return (
            self.temperatureMixinKlasses
            | self.chemistryMixinKlasses
            | self.gasMixinKlasses
            | self.pressureMixinKlasses
            | self.planetMixinKlasses
            | self.starMixinKlasses
            | self.instrumentMixinKlasses
            | self.modelMixinKlasses
            | self.contributionMixinKlasses
            | self.optimizerMixinKlasses
            | self.observationMixinKlasses
        )

    def klass_from_base(self, base_class: AllT) -> t.Set[t.Type[AllT]]:
        """Returns a set of all classes that are subclass of base."""
        return {k for k in self.all_klasses if issubclass(k, base_class)}

    # def suitable_mixins(self, baseclass: AllT) -> t.Set[t.Type[Mixin]]:
    #     """Returns a set of all mixin classes that are suitable for mixin."""
    #     return {m for m in self.mixin_klasses if m.compatible(baseclass)}

    def find_klass(self, name: str) -> t.Optional[t.Union[T, OpacT, MixinT]]:
        """Returns a class with the given name."""
        for klass in self.all_klasses:
            if klass.__name__.lower() == name.lower():
                return klass
        return None

    def find_klass_from_keyword(
        self, keyword: str, baseclass: t.Optional[t.Type[T]] = None
    ) -> t.Type[T]:
        """Returns a class that has the given keyword.

        Returns the first found if multiple klasses have the same keyword.

        """
        klasses = (
            self.all_klasses if baseclass is None else self.klass_from_base(baseclass)
        )

        for k in klasses:
            if hasattr(k, "input_keywords") and keyword.lower() in [
                s.lower() for s in k.input_keywords()
            ]:
                return k
        raise KeyError(f"Could not find class with keyword {keyword}")
